import sys

import CADETMatch.evo as evo
#import grad
import CADETMatch.gradFD as gradFD
import CADETMatch.util as util
import time
import numpy
import shutil
import csv
from cadet import Cadet
from pathlib import Path

from deap import creator
from deap import tools
from deap import base

#parallelization
from scoop import futures
import scoop

import logging

from CADETMatch.cache import cache
import CADETMatch.loggerwriter as loggerwriter
import h5py
import subprocess

#due to how scoop works and the need to break things up into multiple processes it is hard to use class based systems
#As a result most of the code is broken up into modules but is still based on pure functions

#setup scoop logging for all processes

def main(path=None, map_function=None):
    if path is None:
        path = sys.argv[1]
    setup(cache, path, map_function)
    gradFD.setupTemplates(cache)
    #grad.setupTemplates(cache)
    hof = evo.run(cache)

    scoop.logger.info("altFeatures %s", cache.altFeatures)
    scoop.logger.info("altFeatureNames %s", cache.altFeatureNames)

    if cache.altFeatures:
        for name in cache.altFeatureNames:
            json_path = util.setupAltFeature(cache, name)
            setup(cache, json_path, map_function)
            hof = evo.run(cache)

    continue_mcmc(cache, map_function)

    if "repeat" in cache.settings:
        repeat = int(cache.settings['repeat'])

        for i in range(repeat):
            json_path = util.repeatSimulation(i)
            scoop.logger.info(json_path)

            setup(cache, json_path)

            hof = evo.run(cache)
        
        util.metaCSV(cache)

    if "bootstrap" in cache.settings:
        temp = []

        samples = int(cache.settings['bootstrap']['samples'])

        if samples:

            center = float(cache.settings['bootstrap']['center'])
            noise = float(cache.settings['bootstrap']['percentNoise'])/100.0

            bootstrap = cache.settings['resultsDirBase'] / "bootstrap_output"



            for i in range(samples):
                #copy csv files to a new directory with noise added
                #put a new json file in the directory that points to the new csv files
                json_path = util.copyCSVWithNoise(i, center, noise)
                scoop.logger.info(json_path)

                setup(cache, json_path)



                #call setup on all processes with the new json file as an argument to reset them
                #util.updateScores(json_path)

                hof = evo.run(cache)
                temp.append(util.bestMinScore(hof))

                numpy_temp = numpy.array(temp)
                cov = numpy.cov(numpy_temp.transpose())
                scoop.logger.info("in progress cov %s data %s det %s", cov, numpy_temp, numpy.linalg.det(cov))

            numpy_temp = numpy.array(temp)
            cov = numpy.cov(numpy_temp.transpose())
            scoop.logger.info("final cov %s data %s det %s", cov, numpy_temp, numpy.linalg.det(cov))

def setup(cache, json_path, map_function):
    "run seutp for the current json_file"
    cache.setup(json_path)

    createDirectories(cache, json_path)
    createCSV(cache)
    createProgressCSV(cache)
    createErrorCSV(cache)
    setupTemplates(cache)
    setupDeap(cache, map_function)
    setupLog(cache.settings['resultsDirLog'])

def setupLog(log_directory):
    logger = scoop.logger
    logger.propagate = False
    logger.setLevel(logging.INFO)
    # create file handler which logs even debug messages
    fh = logging.FileHandler(log_directory / "main.log")
    fh.setLevel(logging.INFO)

    # add the handlers to the logger
    logger.addHandler(fh)

    sys.stdout = loggerwriter.LoggerWriter(logger.info)
    sys.stderr = loggerwriter.LoggerWriter(logger.warning)

def createDirectories(cache, json_path):
    cache.settings['resultsDirBase'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirGrad'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirMisc'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirSpace'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirEvo'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirProgress'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirMeta'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirLog'].mkdir(parents=True, exist_ok=True)
    cache.settings['resultsDirMCMC'].mkdir(parents=True, exist_ok=True)

    #copy simulation setting file to result base directory
    try:
        shutil.copy(str(json_path), str(cache.settings['resultsDirBase']))
    except shutil.SameFileError:
        pass

def createCSV(cache):
    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    if not path.exists():
        with path.open('w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)
            writer.writerow(cache.headers)

    path = cache.settings['resultsDirMeta'] / 'results.csv'
    if not path.exists():
        with path.open('w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)
            writer.writerow(cache.headers)

def createProgressCSV(cache):
    path = Path(cache.settings['resultsDirBase'], "progress.csv")
    cache.progress_path = path
    if not path.exists():
        with path.open('w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)
            writer.writerow(cache.progress_headers)

def createErrorCSV(cache):
    path = Path(cache.settings['resultsDirBase'], "error.csv")
    cache.error_path = path
    if not path.exists():
        with path.open('w', newline='') as csvfile:
            writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)
            writer.writerow(cache.parameter_headers + ['Return Code', 'STDOUT', 'STDERR'])

def setTemplateValues(simulation, set_values):
    for path, index, value in set_values:
        if index >= 0:
            simulation[path][index] = value
        else:
            simulation[path] = value

def setupTemplates(cache):
    "setup all the experimental templates"
    for experiment in cache.settings['experiments']:
        HDF5 = experiment['HDF5']
        name = experiment['name']

        template_path = Path(cache.settings['resultsDirMisc'], "template_%s.h5" % name)

        template = Cadet()

        #load based on where the HDF5 file is
        template.filename = HDF5
        template.load()

        if 'set_values' in experiment:
            setTemplateValues(template, experiment['set_values'])

        util.setupSimulation(template, cache.target[name]['time'], cache.target[name]['smallest_peak'], cache)

        start = time.time()
        util.runExperiment(None, experiment, cache.settings, cache.target, template, 
                           experiment.get('timeout', 1800), cache)
        elapsed = time.time() - start

        scoop.logger.info("simulation took %s", elapsed)
        
        #timeout needs to be stored in the template so all processes have it without calculating it
        template.root.timeout = elapsed * 10
        
        if cache.settings['searchMethod'] != 'MCMC' and "kde_synthetic" in cache.settings:
            #the base case needs to be saved since the normal template file is what the rest of the code will look for
            template_base_path = Path(cache.settings['resultsDirMisc'], "template_%s_base.h5" % name)
            template.filename = template_base_path.as_posix()
            template.save()
            
            scoop.logger.info("create bias template for experiment %s", name)
            template_bias = util.biasSimulation(template, experiment, cache)
            template_bias_path = Path(cache.settings['resultsDirMisc'], "template_%s_bias.h5" % name)
            template_bias.filename = template_bias_path.as_posix()
            template_bias.save()
            template = template_bias

        #change to where we want the template created
        template.filename = template_path.as_posix()
        
        template.save()

        experiment['simulation'] = template

def setupDeap(cache, map_function):
    "setup the DEAP variables"
    searchMethod = cache.settings.get('searchMethod', 'NSGA3')
    cache.toolbox = base.Toolbox()

    if map_function is None:
        if getattr(scoop, 'SIZE', 1) == 1:
            map_function = map
        else:
            map_function = futures.map

    cache.search[searchMethod].setupDEAP(cache, evo.fitness, gradFD.gradSearch, gradFD.search, map_function, creator, base, tools)

def find_percentile(cache):
    "find the percentile boundaries for the input variables"
    resultDir = Path(cache.settings['resultsDir'])
    result_h5 = resultDir / "result.h5"

    with h5py.File(result_h5, 'r') as hf:
        data = hf["input"][:]
        score = hf["output_meta"][:]

        #product min average norm
        best_norm = numpy.max(score[:,3])

        data = data[score[:,3] > 0.9 * best_norm,:]

        lb, ub = numpy.percentile(data, [0, 100], 0)

        lb_trans = util.convert_individual(lb, cache)[1]
        ub_trans = util.convert_individual(ub, cache)[1]

        scoop.logger.info('lb %s  ub %s', lb, ub)
        scoop.logger.info('lb_trans %s  ub_trans %s', lb_trans, ub_trans)

        return lb_trans, ub_trans

def continue_mcmc(cache, map_function):
    if cache.continueMCMC:
        lb, ub = find_percentile(cache)

        json_path = util.setupMCMC(cache, lb, ub)
        scoop.logger.info(json_path)

        setup(cache, json_path, map_function)

        hof = evo.run(cache)

if __name__ == "__main__":
    start = time.time()
    main()
    scoop.logger.info('System has finished')
    scoop.logger.info("The total runtime was %s seconds" % (time.time() - start))
    sys.exit()


import CADETMatch.util as util

name = "set_value"
count = 0
count_extended = 0

def getUnit(location):
    return location.split('/')[3]

def transform(parameter):
    return []

def untransform(seq, cache, parameter):
    return [], []

def untransform_matrix(matrix, cache, parameter):
    return None

untransform_matrix_inputorder = untransform_matrix

def getValue(sim, location, bound=None, comp=None, index=None):
    if bound is not None:
        unit = getUnit(location)
        boundOffset = util.getBoundOffset(sim.root.input.model[unit])

        if comp == -1:
            position = ()
            return sim[location.lower()]
        else:
            position = boundOffset[comp] + bound
            return sim[location.lower()][position]

    if index is not None:
        if index == -1:
            return sim[location.lower()]
        else:
            return sim[location.lower()][index]

def setValue(sim, value, location, bound=None, comp=None, index=None):
    if bound is not None:
        unit = getUnit(location)
        boundOffset = util.getBoundOffset(sim.root.input.model[unit])

        if comp == -1:
            position = ()
            sim[location.lower()] = value
        else:
            position = boundOffset[comp] + bound
            sim[location.lower()][position] = value

    if index is not None:
        if index == -1:
            sim[location.lower()] = value
        else:
            sim[location.lower()][index] = value
        
def setSimulation(sim, parameter, seq, cache, experiment):
    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:    
        locationFrom = parameter['locationFrom']
        locationTo = parameter['locationTo']
    
        try:
            compFrom = parameter['componentFrom']
            boundFrom = parameter['boundFrom']
            indexFrom = None
        except KeyError:
            indexFrom = parameter['indexFrom']
            boundFrom = None
            compFrom = None
        valueFrom = getValue(sim, locationFrom, bound=boundFrom, comp=compFrom, index=indexFrom)

        try:
            compTo = parameter['componentTo']
            boundTo = parameter['boundTo']
            indexTo = None
        except KeyError:
            indexTo = parameter['indexTo']
            boundTo = None
            compTo = None
        setValue(sim, valueFrom, locationTo, bound=boundTo, comp=compTo, index=indexTo)

    return [],[]

def setupTarget(parameter):
    return [], 0

def getBounds(parameter):
    return None,None

def getHeaders(parameter):
    return []

def getHeadersActual(parameter):
    return []

def setBounds(parameter, lb, ub):
    return None







import CADETMatch.util as util
import random
from pathlib import Path
import csv
import CADETMatch.pareto as pareto
import scoop
import time
import array

name = "Multistart"

def run(cache, tools, creator):
    "run the parameter estimation"
    random.seed()

    parameters = len(cache.MIN_VALUE)

    LAMBDA = parameters * cache.settings['population']
    sim_start = generation_start = time.time()
    result_data = {'input':[], 'output':[], 'output_meta':[], 'results':{}, 'times':{}, 'input_transform':[], 'input_transform_extended':[], 'strategy':[], 
                   'mean':[], 'confidence':[]}

    pop = cache.toolbox.population(n=LAMBDA)

    if "seeds" in cache.settings:
        seed_pop = [cache.toolbox.individual_guess([f(v) for f, v in zip(cache.settings['transform'], sublist)]) for sublist in cache.settings['seeds']]
        pop.extend(seed_pop)

    gradCheck = cache.badScore

    hof = pareto.ParetoFront(similar=util.similar, similar_fit=util.similar_fit)
    meta_hof = pareto.ParetoFront(similar=util.similar, similar_fit=util.similar_fit_meta)
    grad_hof = pareto.ParetoFront(similar=util.similar, similar_fit=util.similar_fit)

    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    with path.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)

        scoop.logger.info("Population %s", [util.convert_individual(i, cache) for i in pop])

        gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, pop, cache, writer, csvfile, hof, meta_hof, -1, check_all=True)

        stalled, stallWarn, progressWarn = util.eval_population(cache.toolbox, cache, newChildren, writer, csvfile, hof, meta_hof, -1, result_data)

        avg, bestMin, bestProd = util.averageFitness(newChildren, cache)
        
        util.writeProgress(cache, -1, newChildren, hof, meta_hof, grad_hof, avg, bestMin, bestProd, sim_start, generation_start, result_data)
        
        util.finish(cache)
        util.graph_corner_process(cache, last=True)

        return hof

def setupDEAP(cache, fitness, grad_fitness, grad_search, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", array.array, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register("individual", util.generateIndividual, creator.Individual,
        len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache)
    
    if cache.sobolGeneration:
        cache.toolbox.register("population", util.sobolGenerator, creator.Individual, cache)
    else:
        cache.toolbox.register("population", tools.initRepeat, list, cache.toolbox.individual)
    cache.toolbox.register("randomPopulation", tools.initRepeat, list, cache.toolbox.individual)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register("evaluate", fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad", grad_fitness, json_path=cache.json_path)
    cache.toolbox.register('grad_search', grad_search)

    cache.toolbox.register('map', map_function)



import CADETMatch.util as util
import sys
import numpy
from addict import Dict

name = "LogSSE"
settings = Dict()
settings.adaptive = False
settings.badScore = -sys.float_info.max
settings.meta_mask = True
settings.count = 1
settings.failure = [0.0] * settings.count, 1e6, 1, numpy.array([0.0]), numpy.array([0.0]), numpy.array([1e6]), [1.0] * settings.count

def run(sim_data, feature):
    "log of SSE score, not composable, negative so score is maximized"
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)
    selected = feature['selected']

    exp_time_values = feature['time'][selected]
    exp_data_values = feature['value'][selected]

    temp = [-numpy.log(util.sse(sim_data_values, exp_data_values)),]

    return (temp, util.sse(sim_data_values, exp_data_values), len(sim_data_values), 
            sim_time_values, sim_data_values, exp_data_values, [-i for i in temp])

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol):
    temp = {}
    temp['peak_max'] = max(selectedValues)
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_SSE" % name]
    return temp



import random
import pickle
import CADETMatch.util as util
import numpy
import scipy
from pathlib import Path
import time
import csv
import cadet

import emcee
import SALib.sample.sobol_sequence

import subprocess
import sys

import CADETMatch.evo as evo
import CADETMatch.cache as cache

import CADETMatch.pareto as pareto

import scoop
import pandas
import array
import CADETMatch.autocorr as autocorr

import CADETMatch.kde_generator as kde_generator
from sklearn.neighbors.kde import KernelDensity

name = "MCMC"

from addict import Dict

import joblib

log2 = numpy.log(2)

def log_previous(cadetValues, kde_previous, kde_previous_scaler):
    #find the right values to use
    col = len(kde_previous_scaler.scale_)
    values = cadetValues[-col:]
    values_shape = numpy.array(values).reshape(1, -1)
    values_scaler = kde_previous_scaler.transform(values_shape)
    score = kde_previous.score_samples(values_scaler)
    return score

def log_likelihood(individual, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup(json_path, False)

    kde_previous, kde_previous_scaler = kde_generator.getKDEPrevious(cache.cache)

    if 'kde' not in log_likelihood.__dict__:
        kde, kde_scaler = kde_generator.getKDE(cache.cache)
        log_likelihood.kde = kde
        log_likelihood.scaler = kde_scaler

    scores, csv_record, results = evo.fitness(individual, json_path)

    if results is None:
        return -numpy.inf, scores, csv_record, results

    if results is not None and kde_previous is not None:
        logPrevious = log_previous(individual, kde_previous, kde_previous_scaler)
    else:
        logPrevious = 0.0

    scores_shape = numpy.array(scores).reshape(1, -1)

    score_scaler = log_likelihood.scaler.transform(scores_shape)

    score_kde = log_likelihood.kde.score_samples(score_scaler)

    score = score_kde + log2 + logPrevious #*2 is from mirroring and we need to double the probability to get back to the normalized distribution

    return score, scores, csv_record, results 

def log_posterior(theta, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup(json_path)

    #try:
    ll, scores, csv_record, results = log_likelihood(theta, json_path)
    if results is None:
        return -numpy.inf, None, None, None, None
    else:
        return ll, theta, scores, csv_record, results
    #except:
    #    # if model does not converge:
    #    return -numpy.inf, None, None, None, None

def addChain(*args):
    temp = [arg for arg in args if arg is not None]
    if len(temp) > 1:
        return numpy.concatenate( temp, axis=1)
    else:
        return numpy.array(temp[0])

def flatten(chain):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])
    return flat_chain

def sampler_burn(cache, checkpoint, sampler, checkpointFile):
    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)

    train_chain_stat = checkpoint.get('train_chain_stat', None)
    run_chain_stat = checkpoint.get('run_chain_stat', None)

    converge = checkpoint.get('converge')    

    parameters = len(cache.MIN_VALUE)

    tol = 5e-4
    power = 0.0
    distance = 1.0
    distance_a = sampler.a
    stop_next = False
    finished = False

    generation = checkpoint['idx_burn']

    sampler.iterations = checkpoint['sampler_iterations']
    sampler.naccepted = checkpoint['sampler_naccepted']
    sampler.a = checkpoint['sampler_a']

    while not finished:
        p, ln_prob, random_state = next(sampler.sample(checkpoint['p_burn'], lnprob0=checkpoint['ln_prob_burn'], rstate0=checkpoint['rstate_burn'], iterations=1 ))

        accept = numpy.mean(sampler.acceptance_fraction)
        burn_seq.append(accept)
        converge[:-1] = converge[1:]
        converge[-1] = accept

        train_chain = addChain(train_chain, p[:, numpy.newaxis, :])

        train_chain_stat = addChain(train_chain_stat, numpy.percentile(flatten(train_chain), [5, 50, 95], 0)[:, numpy.newaxis, :])

        converge_real = converge[~numpy.isnan(converge)]
        scoop.logger.info('burn:  idx: %s accept: %.3g std: %.3g mean: %.3g converge: %.3g', generation, accept, 
                            numpy.std(converge_real), numpy.mean(converge_real), numpy.std(converge_real)/tol)

        generation += 1

        checkpoint['p_burn'] = p
        checkpoint['ln_prob_burn'] = ln_prob
        checkpoint['rstate_burn'] = random_state
        checkpoint['idx_burn'] = generation
        checkpoint['train_chain'] = train_chain
        checkpoint['burn_seq'] = burn_seq
        checkpoint['converge'] = converge
        checkpoint['sampler_iterations'] = sampler.iterations
        checkpoint['sampler_naccepted'] = sampler.naccepted
        checkpoint['train_chain_stat'] = train_chain_stat
        checkpoint['run_chain_stat'] = run_chain_stat
        checkpoint['sampler_a'] = sampler.a

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, train_chain, run_chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, None)
        util.graph_corner_process(cache, last=False)

        if numpy.std(converge_real) < tol and len(converge) == len(converge_real):
            average_converge = numpy.mean(converge)
            if 0.16 < average_converge < 0.30:
                scoop.logger.info("burn in completed at iteration %s", generation)
                finished = True

            if stop_next is True:
                scoop.logger.info("burn in completed at iteration %s based on minimum distances", generation)
                finished = True

            if not finished:
                new_distance = numpy.abs(average_converge - 0.234)
                if new_distance < distance:
                    distance = new_distance
                    distance_a = sampler.a

                    scoop.logger.info("burn in acceptance is out of tolerance and alpha must be adjusted while burn in continues")
                    converge[:] = numpy.nan
                    prev_a = sampler.a
                    if average_converge > 0.3:
                        #a must be increased to decrease the acceptance rate (step size)
                        power += 1
                    else:
                        #a must be decreased to increase the acceptance rate (step size)
                        power -= 1
                    new_a = 1.0 + 2.0*power
                    sampler.a = new_a
                    sampler.reset()
                    scoop.logger.info('previous alpha: %s    new alpha: %s', prev_a, new_a)
                else:
                    sampler.a = distance_a
                    sampler.reset()
                    stop_next = True
 
    sampler.reset()

    checkpoint['sampler_iterations'] = sampler.iterations
    checkpoint['sampler_naccepted'] = sampler.naccepted
    checkpoint['state'] = 'chain'
    checkpoint['p_chain'] = p
    checkpoint['ln_prob_chain'] = None
    checkpoint['rstate_chain'] = None
    checkpoint['sampler_a'] = sampler.a

    write_interval(-1, cache, checkpoint, checkpointFile, train_chain, run_chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, None)
            

def sampler_run(cache, checkpoint, sampler, checkpointFile):
    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)

    train_chain_stat = checkpoint.get('train_chain_stat', None)
    run_chain_stat = checkpoint.get('run_chain_stat', None)

    checkInterval = 25

    parameters = len(cache.MIN_VALUE)

    finished = False

    generation = checkpoint['idx_chain']

    sampler.iterations = checkpoint['sampler_iterations']
    sampler.naccepted = checkpoint['sampler_naccepted']
    sampler.a = checkpoint['sampler_a']
    tau_percent = None

    while not finished:
        p, ln_prob, random_state = next(sampler.sample(checkpoint['p_chain'], lnprob0=checkpoint['ln_prob_chain'], rstate0=checkpoint['rstate_chain'], iterations=1 ))

        accept = numpy.mean(sampler.acceptance_fraction)
        chain_seq.append(accept)

        run_chain = addChain(run_chain, p[:, numpy.newaxis, :])

        run_chain_stat = addChain(run_chain_stat, numpy.percentile(flatten(run_chain), [5, 50, 95], 0)[:, numpy.newaxis, :])

        scoop.logger.info('run:  idx: %s accept: %.3f', generation, accept)
        
        generation += 1

        checkpoint['p_chain'] = p
        checkpoint['ln_prob_chain'] = ln_prob
        checkpoint['rstate_chain'] = random_state
        checkpoint['idx_chain'] = generation
        checkpoint['run_chain'] = run_chain
        checkpoint['chain_seq'] = chain_seq
        checkpoint['sampler_iterations'] = sampler.iterations
        checkpoint['sampler_naccepted'] = sampler.naccepted
        checkpoint['train_chain_stat'] = train_chain_stat
        checkpoint['run_chain_stat'] = run_chain_stat

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, train_chain, run_chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent)
        mle_process(last=False)
        util.graph_corner_process(cache, last=False)

        if generation % checkInterval == 0:
            try:
                tau = autocorr.integrated_time(numpy.swapaxes(run_chain, 0, 1), tol=cache.MCMCTauMult)
                scoop.logger.info("Mean acceptance fraction: %s %0.3f tau: %s with shape: %s", generation, accept, tau, run_chain.shape)
                if numpy.any(numpy.isnan(tau)):
                    scoop.logger.info("tau is NaN and clearly not complete %s", generation)
                else:
                    scoop.logger.info("we have run long enough and can quit %s", generation)
                    finished = True
            except autocorr.AutocorrError as err:
                scoop.logger.info(str(err))
                tau = err.tau
            scoop.logger.info("Mean acceptance fraction: %s %0.3f tau: %s", generation, accept, tau)
            tau = numpy.array(tau)
            tau_percent = generation / (tau * cache.MCMCTauMult)

    checkpoint['p_chain'] = p
    checkpoint['ln_prob_chain'] = ln_prob
    checkpoint['rstate_chain'] = random_state
    checkpoint['idx_chain'] = generation
    checkpoint['run_chain'] = run_chain
    checkpoint['chain_seq'] = chain_seq
    checkpoint['state'] = 'complete'

    write_interval(-1, cache, checkpoint, checkpointFile, train_chain, run_chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent)

def write_interval(interval, cache, checkpoint, checkpointFile, train_chain, run_chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent=None):
    "write the checkpoint and mcmc data at most every n seconds"
    if 'last_time' not in write_interval.__dict__:
        write_interval.last_time = time.time()

    if time.time() - write_interval.last_time > interval:
        with checkpointFile.open('wb') as cp_file:
            pickle.dump(checkpoint, cp_file)

        writeMCMC(cache, train_chain, run_chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent)
        
        write_interval.last_time = time.time()

def run(cache, tools, creator):
    "run the parameter estimation"
    random.seed()
    checkpointFile = Path(cache.settings['resultsDirMisc'], cache.settings.get('checkpointFile', 'check'))
    checkpoint = getCheckPoint(checkpointFile,cache)

    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)

    parameters = len(cache.MIN_VALUE)
    
    MCMCpopulationSet = cache.settings.get('MCMCpopulationSet', None)
    if MCMCpopulationSet is not None:
        populationSize = MCMCpopulationSet
    else:
        populationSize = parameters * cache.settings['MCMCpopulation']
               
    #Population must be even
    populationSize = populationSize + populationSize % 2  

    if checkpoint['state'] == 'start':
        scoop.logger.info("setting up kde")
        kde, kde_scaler = kde_generator.setupKDE(cache)
        checkpoint['state'] = 'burn_in'
    else:
        scoop.logger.info("loading kde")
        kde, kde_scaler = kde_generator.getKDE(cache)

    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    with path.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)

        sampler = emcee.EnsembleSampler(populationSize, parameters, log_posterior, args=[cache.json_path], pool=cache.toolbox, a=2.0)
        emcee.EnsembleSampler._get_lnprob = _get_lnprob
        emcee.EnsembleSampler._propose_stretch = _propose_stretch

        if 'sampler_a' not in checkpoint:
            checkpoint['sampler_a'] = sampler.a

        result_data = {'input':[], 'output':[], 'output_meta':[], 'results':{}, 'times':{}, 'input_transform':[], 'input_transform_extended':[], 'strategy':[], 
                   'mean':[], 'confidence':[], 'mcmc_score':[]}
        halloffame = pareto.DummyFront(similar=util.similar)
        meta_hof = pareto.ParetoFront(similar=util.similar, similar_fit=util.similar_fit_meta)
        grad_hof = pareto.DummyFront(similar=util.similar)

        def local(results):
            return process(cache, halloffame, meta_hof, grad_hof, result_data, results, writer, csvfile, sampler)
        sampler.process = local

        if checkpoint['state'] == 'burn_in':
            sampler_burn(cache, checkpoint, sampler, checkpointFile)

        if checkpoint['state'] == 'chain':
            run_chain = checkpoint.get('run_chain', None)
            if run_chain is not None:
                temp = run_chain[:, :checkpoint['idx_chain'], 0].T
                scoop.logger.info('complete shape %s', temp.shape)
 
                try:
                    tau = autocorr.integrated_time(numpy.swapaxes(run_chain[:, :checkpoint['idx_chain'], :], 0, 1), tol=cache.MCMCTauMult)
                    scoop.logger.info("we have previously run long enough and can quit %s", checkpoint['idx_chain'])
                    checkpoint['state'] = 'complete'
                except autocorr.AutocorrError as err:
                    scoop.logger.info(str(err))
                    tau = err.tau

                scoop.logger.info("Mean acceptance fraction: %s %0.3f tau: %s", checkpoint['idx_chain'], checkpoint['chain_seq'][-1], tau)
            
        if checkpoint['state'] == 'chain':
            sampler_run(cache, checkpoint, sampler, checkpointFile)

    chain = checkpoint['run_chain']
    chain_shape = chain.shape
    chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])

    if checkpoint['state'] == 'complete':
        tube_process(last=True)
        util.finish(cache)
        checkpoint['state'] = 'plot_finish'

        with checkpointFile.open('wb') as cp_file:
            pickle.dump(checkpoint, cp_file)

    if checkpoint['state'] == 'plot_finish':
        mle_process(last=True)
        util.graph_corner_process(cache, last=True)
    return numpy.mean(chain, 0)

def tube_process(last=False, interval=3600):
    cwd = str(Path(__file__).parent.parent)
    ret = subprocess.run([sys.executable, '-m', 'scoop', 'mcmc_plot_tube.py', str(cache.cache.json_path),], 
            stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)

def mle_process(last=False, interval=3600):
    if 'last_time' not in mle_process.__dict__:
        mle_process.last_time = time.time()

    if 'child' in mle_process.__dict__:
        if mle_process.child.poll() is None:  #This is false if the child has completed
            if last:
                mle_process.child.wait()
            else:
                return

    cwd = str(Path(__file__).parent.parent)

    if last:
        ret = subprocess.run([sys.executable, '-m', 'scoop', 'mle.py', str(cache.cache.json_path),], 
            stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)
        mle_process.last_time = time.time()
    elif (time.time() - mle_process.last_time) > interval:
        #mle_process.child = subprocess.Popen([sys.executable, '-m', 'scoop', 'mle.py', str(cache.cache.json_path),], 
        #    stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)
        subprocess.run([sys.executable, '-m', 'scoop', 'mle.py', str(cache.cache.json_path),], 
            stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)
        mle_process.last_time = time.time()
        
def get_population(base, size, diff=0.05):
    new_population = base
    row, col = base.shape
    scoop.logger.info('%s', base)
    
    scoop.logger.info('row %s size %s', row, size)
    if row < size:
        #create new entries
        indexes = numpy.random.choice(new_population.shape[0], size - row, replace=True)
        temp = new_population[indexes,:]
        rand = numpy.random.uniform(1.0-diff, 1.0+diff, size=temp.shape)
        new_population = numpy.concatenate([new_population, temp * rand])
    if row > size:
        #randomly select entries to keep
        indexes = numpy.random.choice(new_population.shape[0], size, replace=False)
        scoop.logger.info('indexes: %s', indexes)
        new_population = new_population[indexes,:]
    return new_population

def getCheckPoint(checkpointFile, cache):
    if checkpointFile.exists():
        with checkpointFile.open('rb') as cp_file:
            checkpoint = pickle.load(cp_file)
    else:
        parameters = len(cache.MIN_VALUE)
        
        MCMCpopulationSet = cache.settings.get('MCMCpopulationSet', None)
        if MCMCpopulationSet is not None:
            populationSize = MCMCpopulationSet
        else:
            populationSize = parameters * cache.settings['MCMCpopulation']

        #Population must be even
        populationSize = populationSize + populationSize % 2  

        checkpoint = {}
        checkpoint['state'] = 'start'

        if cache.settings.get('PreviousResults', None) is not None:
            scoop.logger.info('running with previous best results')
            previousResultsFile = Path(cache.settings['PreviousResults'])
            results_h5 = cadet.H5()
            results_h5.filename = previousResultsFile.as_posix()
            results_h5.load()
            previousResults = results_h5.root.meta_population_transform

            row,col = previousResults.shape
            scoop.logger.info('row: %s col: %s  parameters: %s', row, col, parameters)
            if col < parameters:
                mcmc_h5 = Path(cache.settings.get('mcmc_h5', None))
                mcmcDir = mcmc_h5.parent
                mle_h5 = mcmcDir / "mle.h5"

                data = cadet.H5()
                data.filename = mle_h5.as_posix()
                data.load()
                scoop.logger.info('%s', list(data.root.keys()))
                stat_MLE = data.root.stat_MLE.reshape(1, -1)
                previousResults = numpy.hstack([previousResults, numpy.repeat(stat_MLE, row, 0)])
                scoop.logger.info('row: %s  col:%s   shape: %s', row, col, previousResults.shape)

            population = get_population(previousResults, populationSize, diff=0.1)
            checkpoint['p_burn'] = [util.convert_individual_inverse(i, cache) for i in population]
            scoop.logger.info('p_burn startup: %s', checkpoint['p_burn'])
        else:
            checkpoint['p_burn'] = SALib.sample.sobol_sequence.sample(populationSize, parameters)

        checkpoint['ln_prob_burn'] = None
        checkpoint['rstate_burn'] = None
        checkpoint['idx_burn'] = 0
        
        checkpoint['p_chain'] = None
        checkpoint['ln_prob_chain'] = None
        checkpoint['rstate_chain'] = None
        checkpoint['idx_chain'] = 0

        checkpoint['sampler_iterations'] = 0
        checkpoint['sampler_naccepted'] = numpy.zeros(populationSize)

        checkpoint['converge'] = numpy.ones(cache.settings.get('burnStable', 50)) * numpy.nan
    
    checkpoint['length_chain'] = cache.settings.get('chainLength', 50000)
    checkpoint['length_burn'] = cache.settings.get('burnIn', 50000)
    return checkpoint

def setupDEAP(cache, fitness, grad_fitness, grad_search, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", array.array, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register("individual", util.generateIndividual, creator.Individual,
        len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register('map', map_function)

def process(cache, halloffame, meta_hof, grad_hof, result_data, results, writer, csv_file, sampler):
    if 'gen' not in process.__dict__:
        process.gen = 0

    if 'sim_start' not in process.__dict__:
        process.sim_start = time.time()

    if 'generation_start' not in process.__dict__:
        process.generation_start = time.time()

    population = []
    fitnesses = []
    for sse, theta, fit, csv_line, result in results:
        if result is not None:
            parameters = theta
            fitnesses.append( (fit, csv_line, result) )

            ind = cache.toolbox.individual_guess(parameters)
            population.append(ind)
            result_data['mcmc_score'].append(sse)

    stalled, stallWarn, progressWarn = util.process_population(cache.toolbox, cache, population, 
                                                          fitnesses, writer, csv_file, 
                                                          halloffame, meta_hof, process.gen, result_data)
    
    avg, bestMin, bestProd = util.averageFitness(population, cache)

    util.writeProgress(cache, process.gen, population, halloffame, meta_hof, grad_hof, avg, bestMin, bestProd, 
                       process.sim_start, process.generation_start, result_data, line_log=False)

    util.graph_process(cache, process.gen)

    process.gen += 1
    process.generation_start = time.time()
    return [i[0] for i in results]

def _get_lnprob(self, pos=None):
    """
    Calculate the vector of log-probability for the walkers.

    :param pos: (optional)
        The position vector in parameter space where the probability
        should be calculated. This defaults to the current position
        unless a different one is provided.

    This method returns:

    * ``lnprob`` - A vector of log-probabilities with one entry for each
        walker in this sub-ensemble.

    * ``blob`` - The list of meta data returned by the ``lnpostfn`` at
        this position or ``None`` if nothing was returned.

    """
    if pos is None:
        p = self.pos
    else:
        p = pos

    # Check that the parameters are in physical ranges.
    if numpy.any(numpy.isinf(p)):
        raise ValueError("At least one parameter value was infinite.")
    if numpy.any(numpy.isnan(p)):
        raise ValueError("At least one parameter value was NaN.")

    # If the `pool` property of the sampler has been set (i.e. we want
    # to use `multiprocessing`), use the `pool`'s map method. Otherwise,
    # just use the built-in `map` function.
    if self.pool is not None:
        M = self.pool.map
    else:
        M = map

    # sort the tasks according to (user-defined) some runtime guess
    if self.runtime_sortingfn is not None:
        p, idx = self.runtime_sortingfn(p)

    # Run the log-probability calculations (optionally in parallel).
    results = list(M(self.lnprobfn, [p[i] for i in range(len(p))]))
    results = self.process(results)

    try:
        lnprob = numpy.array([float(l[0]) for l in results])
        blob = [l[1] for l in results]
    except (IndexError, TypeError):
        lnprob = numpy.array([float(l) for l in results])
        blob = None

    # sort it back according to the original order - get the same
    # chain irrespective of the runtime sorting fn
    if self.runtime_sortingfn is not None:
        orig_idx = numpy.argsort(idx)
        lnprob = lnprob[orig_idx]
        p = [p[i] for i in orig_idx]
        if blob is not None:
            blob = [blob[i] for i in orig_idx]

    # Check for lnprob returning NaN.
    if numpy.any(numpy.isnan(lnprob)):
        # Print some debugging stuff.
        bad_pars = []
        for pars in p[numpy.isnan(lnprob)]:
            bad_pars.append(pars)
        scoop.logger.info("NaN value of lnprob for parameters: %s", bad_pars)

        # Finally raise exception.
        raise ValueError("lnprob returned NaN.")

    return lnprob, blob

def process_chain(chain, cache, idx):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])

    flat_chain_transform = util.convert_population(flat_chain, cache)
    chain_transform = flat_chain_transform.reshape(chain_shape)

    return chain, flat_chain, chain_transform, flat_chain_transform

def writeMCMC(cache, train_chain, chain, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent):
    "write out the mcmc data so it can be plotted"
    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    train_chain, train_chain_flat, train_chain_transform, train_chain_flat_transform = process_chain(train_chain, cache, len(burn_seq)-1)

    if chain is not None:
        chain, chain_flat, chain_transform, chain_flat_transform = process_chain(chain, cache, len(chain_seq)-1)
        interval_chain = chain_flat
        interval_chain_transform = chain_flat_transform
    else:
        interval_chain = train_chain_flat
        interval_chain_transform = train_chain_flat_transform

    flat_interval = interval(interval_chain, cache)
    flat_interval_transform = interval(interval_chain_transform, cache)

    flat_interval.to_csv(mcmcDir / "percentile.csv")
    flat_interval_transform.to_csv(mcmcDir / "percentile_transform.csv")

    h5 = cadet.H5()
    h5.filename = mcmc_h5.as_posix()

    if tau_percent is not None:
        h5.root.tau_percent = tau_percent.reshape(-1, 1)

    if train_chain_stat is not None:
        train_chain_stat, _, train_chain_stat_transform, _ = process_chain(train_chain_stat, cache, len(burn_seq)-1)

        h5.root.train_chain_stat = train_chain_stat
        h5.root.train_chain_stat_transform = train_chain_stat_transform

    if run_chain_stat is not None:
        run_chain_stat, _, run_chain_stat_transform, _ = process_chain(run_chain_stat, cache, len(burn_seq)-1)

        h5.root.run_chain_stat = run_chain_stat
        h5.root.run_chain_stat_transform = run_chain_stat_transform

    if burn_seq:
        h5.root.burn_in_acceptance = numpy.array(burn_seq).reshape(-1, 1)

    if chain_seq:
        h5.root.mcmc_acceptance = numpy.array(chain_seq).reshape(-1, 1)

    if chain is not None:
        h5.root.full_chain = chain
        h5.root.full_chain_transform = chain_transform
        h5.root.flat_chain = chain_flat
        h5.root.flat_chain_transform = chain_flat_transform

    h5.root.train_full_chain = train_chain
    h5.root.train_full_chain_transform = train_chain_transform
    h5.root.train_flat_chain = train_chain_flat
    h5.root.train_flat_chain_transform = train_chain_flat_transform

    mean = numpy.mean(interval_chain_transform,0)
    labels = [5, 10, 50, 90, 95]
    percentile = numpy.percentile(interval_chain_transform, labels, 0)

    h5.root.percentile['mean'] = mean
    for idx, label in enumerate(labels):
        h5.root.percentile['percentile_%s' % label] = percentile[idx,:]
        
    h5.save()

def interval(flat_chain, cache):
    mean = numpy.mean(flat_chain,0)

    percentile = numpy.percentile(flat_chain, [5, 10, 50, 90, 95], 0)

    data = numpy.vstack( (mean, percentile) ).transpose()

    pd = pandas.DataFrame(data, columns = ['mean', '5', '10', '50', '90', '95'])
    pd.insert(0, 'name', cache.parameter_headers_actual)
    pd.set_index('name')
    return pd

def _propose_stretch(self, p0, p1, lnprob0):
    """
    Propose a new position for one sub-ensemble given the positions of
    another.

    :param p0:
        The positions from which to jump.

    :param p1:
        The positions of the other ensemble.

    :param lnprob0:
        The log-probabilities at ``p0``.

    This method returns:

    * ``q`` - The new proposed positions for the walkers in ``ensemble``.

    * ``newlnprob`` - The vector of log-probabilities at the positions
        given by ``q``.

    * ``accept`` - A vector of type ``bool`` indicating whether or not
        the proposed position for each walker should be accepted.

    * ``blob`` - The new meta data blobs or ``None`` if nothing was
        returned by ``lnprobfn``.

    """
    s = numpy.atleast_2d(p0)
    Ns = len(s)
    c = numpy.atleast_2d(p1)
    Nc = len(c)

    # Generate the vectors of random numbers that will produce the
    # proposal.
    zz = ((self.a - 1.) * self._random.rand(Ns) + 1) ** 2. / self.a
    rint = self._random.randint(Nc, size=(Ns,))

    # Calculate the proposed positions and the log-probability there.
    q = (c[rint] - zz[:, numpy.newaxis] * (c[rint] - s)) % 1
        
    newlnprob, blob = self._get_lnprob(q)

    # Decide whether or not the proposals should be accepted.
    lnpdiff = (self.dim - 1.) * numpy.log(zz) + newlnprob - lnprob0
    accept = (lnpdiff > numpy.log(self._random.rand(len(lnpdiff))))

    return q, newlnprob, accept, blob


#This file is from https://raw.githubusercontent.com/lmarti/nsgaiii/master/nsgaiii/selection.py


#    This file is part of nsgaiii, a Python implementation of NSGA-III.
#
#    nsgaiii is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    nsgaiii is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with nsgaiii. If not, see <http://www.gnu.org/licenses/>.
#
#    by Luis Marti (IC/UFF) http://lmarti.com

from __future__ import division # making it work with Python 2.x
import random
import numpy as np
from deap import tools

import SALib.sample.sobol_sequence

class ReferencePoint(list):
    '''A reference point exists in objective space an has a set of individuals
    associated to it.'''
    def __init__(self, *args):
        list.__init__(self, *args)
        self.associations_count = 0
        self.associations = []

def generate_reference_points(num_objs, num_divisions_per_obj=4):
    '''Generates reference points for NSGA-III selection. This code is based on
    `jMetal NSGA-III implementation <https://github.com/jMetal/jMetal>`_.
    '''
    sobol = SALib.sample.sobol_sequence.sample(100 * num_objs, num_objs)
    data = np.apply_along_axis(list, 1, sobol)
    data = list(map(ReferencePoint, data))
    return data
    #This entire function needs to be completely recoded, in high dimensions in it is major time sink
    # 287/622s. This should probably be done once and stored instead of regenerated at every iteration
    #def gen_refs_recursive(work_point, num_objs, left, total, depth):
    #    if depth == num_objs - 1:
    #        work_point[depth] = left/total
    #        ref = ReferencePoint(copy.deepcopy(work_point))
    #        return [ref]
    #    else:
    #        res = []
    #        for i in range(left):
    #            work_point[depth] = i/total
    #            res = res + gen_refs_recursive(work_point, num_objs, left-i, total, depth+1)
    #        return res
    #return gen_refs_recursive([0]*num_objs, num_objs, num_objs*num_divisions_per_obj,
    #                          num_objs*num_divisions_per_obj, 0)

def find_ideal_point(individuals):
    'Finds the ideal point from a set individuals.'
    current_ideal = [np.infty] * len(individuals[0].fitness.values)
    for ind in individuals:
        # Use wvalues to accomodate for maximization and minimization problems.
        current_ideal = np.minimum(current_ideal,
                                   np.multiply(ind.fitness.wvalues, -1))
    return current_ideal

def find_extreme_points(individuals):
    'Finds the individuals with extreme values for each objective function.'
    return [sorted(individuals, key=lambda ind: ind.fitness.wvalues[o] * -1)[-1]
            for o in range(len(individuals[0].fitness.values))]

def construct_hyperplane(individuals, extreme_points):
    'Calculates the axis intersects for a set of individuals and its extremes.'
    def has_duplicate_individuals(individuals):
        for i in range(len(individuals)):
            for j in range(i+1, len(individuals)):
                if individuals[i].fitness.values == individuals[j].fitness.values:
                    return True
        return False

    num_objs = len(individuals[0].fitness.values)

    if has_duplicate_individuals(extreme_points):
        intercepts = [extreme_points[m].fitness.values[m] for m in range(num_objs)]
    else:
        b = np.ones(num_objs)
        A = [point.fitness.values for point in extreme_points]
        x = np.linalg.solve(A, b)
        intercepts = 1/x
    return intercepts

def normalize_objective(individual, m, intercepts, ideal_point, epsilon=1e-20):
    'Normalizes an objective.'
    # Numeric trick present in JMetal implementation.
    if np.abs(intercepts[m]-ideal_point[m] > epsilon):
        return individual.fitness.values[m] / (intercepts[m]-ideal_point[m])
    else:
        return individual.fitness.values[m] / epsilon

def normalize_objectives(individuals, intercepts, ideal_point):
    '''Normalizes individuals using the hyperplane defined by the intercepts as
    reference. Corresponds to Algorithm 2 of Deb & Jain (2014).'''
    num_objs = len(individuals[0].fitness.values)

    for ind in individuals:
        ind.fitness.normalized_values = list([normalize_objective(ind, m,
                                                                  intercepts, ideal_point)
                                                                  for m in range(num_objs)])
    return individuals

def perpendicular_distance(direction, point):
    k = np.dot(direction, point) / np.sum(np.power(direction, 2))
    d = np.sum(np.power(np.subtract(np.multiply(direction, [k] * len(direction)), point) , 2))
    return np.sqrt(d)

def perpendicular_distances(direction, points):
    #This function is also slow but I am not sure what else to do with it
    #Need to test with numpy connected to MKL
    t = np.dot(points,direction)/np.linalg.norm(direction)**2
    d = np.linalg.norm(points - np.outer(t, direction), axis=1)
    return d

def associate(individuals, reference_points):
    '''Associates individuals to reference points and calculates niche number.
    Corresponds to Algorithm 3 of Deb & Jain (2014).'''
    pareto_fronts = tools.sortLogNondominated(individuals, len(individuals))
    num_objs = len(individuals[0].fitness.values)

    for ind in individuals:
        dists = perpendicular_distances(ind.fitness.normalized_values, reference_points)
        best_idx = np.argmin(dists)
        best_rp = reference_points[best_idx]
        best_dist = dists[best_idx]

        ind.reference_point = best_rp
        ind.ref_point_distance = best_dist
        best_rp.associations_count += 1 # update de niche number
        best_rp.associations += [ind]

def niching_select(individuals, k):
    '''Secondary niched selection based on reference points. Corresponds to
    steps 13-17 of Algorithm 1 and to Algorithm 4.'''
    if len(individuals) == k:
        return individuals

    #individuals = copy.deepcopy(individuals)

    ideal_point = find_ideal_point(individuals)
    extremes = find_extreme_points(individuals)
    intercepts = construct_hyperplane(individuals, extremes)
    normalize_objectives(individuals, intercepts, ideal_point)

    reference_points = generate_reference_points(len(individuals[0].fitness.values))

    ref = np.array(reference_points)

    associate(individuals, reference_points)

    res = []
    while len(res) < k:
        #This can still be made more efficient by updating the associations count when it changes during iteration
        #instead of rebuilding the system. Right now this takes no time at all anymore when profiling
        #unless it takes more time in higher dimensions or larger problems there is no point in further
        #optimization here

        rp_associations_count = np.fromiter((rp.associations_count for rp in reference_points), int, len(reference_points))

        min_assoc_rp = np.min(rp_associations_count[np.nonzero(rp_associations_count)])

        idxs = np.where(rp_associations_count == min_assoc_rp)

        chosen_rp = reference_points[random.choice(idxs[0])]

        associated_inds = chosen_rp.associations
        if chosen_rp.associations:
            if chosen_rp.associations_count == 0:
                sel = min(chosen_rp.associations, key=lambda ind: ind.ref_point_distance)
            else:
                sel = chosen_rp.associations[random.randint(0, len(chosen_rp.associations)-1)]
            res += [sel]
            chosen_rp.associations.remove(sel)
            chosen_rp.associations_count += 1
            individuals.remove(sel)
        else:
            reference_points.remove(chosen_rp)
    
    #These attributes are no longer needed but make the checkpoints huge        
    for i in res:
        i.reference_point = None
        i.ref_point_distance = None

    return res

def sel_nsga_iii(individuals, k):
    '''Implements NSGA-III selection as described in
    Deb, K., & Jain, H. (2014). An Evolutionary Many-Objective Optimization
    Algorithm Using Reference-Point-Based Nondominated Sorting Approach,
    Part I: Solving Problems With Box Constraints. IEEE Transactions on
    Evolutionary Computation, 18(4), 577–601. doi:10.1109/TEVC.2013.2281535.
    '''
    assert len(individuals) >= k

    if len(individuals) == k:
        return individuals

    # Algorithm 1 steps 4--8
    fronts = tools.sortLogNondominated(individuals, len(individuals))

    limit = 0
    res = []
    for f, front in enumerate(fronts):
        res += front
        if len(res) > k:
            limit = f
            break
    # Algorithm 1 steps
    selection = []
    if limit > 0:
        for f in range(limit):
            selection += fronts[f]

    # complete selected inividuals using the referece point based approach
    selection += niching_select(fronts[limit], k - len(selection))
    return selection

__all__ = ["sel_nsga_iii"]



import sys

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from cadet import H5

from CADETMatch.cache import cache

from pathlib import Path

def main():
    cache.setup(sys.argv[1])
    
    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.load()

    store = h5_data.root.store

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    plt.figure(figsize=[10,10])
    plt.scatter(numpy.log10(store[:,0]), numpy.log10(store[:,1]))
    plt.xlabel('bandwidth')
    plt.ylabel('cross_val_score')
    plt.savefig(str(mcmcDir / "log_bandwidth.png" ), bbox_inches='tight')
    plt.close()

    plt.figure(figsize=[10,10])
    plt.scatter(store[:,0], store[:,1])
    plt.xlabel('bandwidth')
    plt.ylabel('cross_val_score')
    plt.savefig(str(mcmcDir / "bandwidth.png" ), bbox_inches='tight')
    plt.close()

if __name__ == "__main__":
    main()





import CADETMatch.util as util
import CADETMatch.score as score
import numpy
import pandas
import scipy.optimize
from addict import Dict

name = "fractionationSlide"
settings = Dict()
settings.adaptive = True
settings.badScore = 0
settings.meta_mask = True
settings.count = None


def goal(offset, frac_exp, sim_data_time, sim_data_value, start, stop):
    sim_data_value = score.roll_spline(sim_data_time, sim_data_value, offset)
    frac_sim = util.fractionate(start, stop, sim_data_time, sim_data_value)
    return numpy.sum((frac_exp-frac_sim)**2)

def searchRange(times, start_frac, stop_frac, CV_time):
    collectionStart = min(start_frac)
    collectionStop = max(stop_frac)

    searchStart = collectionStart - CV_time
    searchStop = collectionStop + CV_time

    searchStart = max(searchStart, times[0])
    searchStop = min(searchStop, times[-1])

    searchIndexStart = numpy.argmax( times[times <= searchStart])
    searchIndexStop = numpy.argmax( times[times <= searchStop])
    return searchIndexStart, searchIndexStop

def run(sim_data, feature):
    simulation = sim_data['simulation']
    timeFunc = feature['timeFunc']
    components = feature['components']
    numComponents = len(components)
    samplesPerComponent = feature['samplesPerComponent']
    data = feature['data']
    CV_time = feature['CV_time']
    start = feature['start']
    stop = feature['stop']
    funcs = feature['funcs']

    time_center = (start + stop)/2.0

    times = simulation.root.output.solution.solution_times

    scores = []

    sim_values_sse = []
    exp_values_sse = []
   
    graph_sim = {}
    graph_exp = {}

    searchIndexStart, searchIndexStop = searchRange(times, start, stop, CV_time)

    for component, value_func in funcs:
        exp_values = numpy.array(data[str(component)])
        sim_value = simulation.root.output.solution[feature['unit']]["solution_outlet_comp_%03d" % component]

        bounds = find_bounds(times, sim_value)
        result = scipy.optimize.differential_evolution(goal, bounds = [bounds,], 
                                                       args = (exp_values, times, sim_value, start, stop))

        time_offset = abs(result.x[0])
        sim_data_value = score.roll_spline(times, sim_value, result.x[0])

        fracOffset = util.fractionate(start, stop, times, sim_data_value)

        value_score = value_func(max(fracOffset))
        pear = score.pear_corr(scipy.stats.pearsonr(exp_values, fracOffset)[0])
        time_score = timeFunc(time_offset)

        exp_values_sse.extend(exp_values)
        sim_values_sse.extend(fracOffset)

        scores.append(pear)
        scores.append(time_score)
        scores.append(value_score)

        graph_sim[component] = list(zip(time_center, fracOffset))
        graph_exp[component] = list(zip(time_center, exp_values))

    sim_data['graph_exp'] = graph_exp
    sim_data['graph_sim'] = graph_sim

    return (scores, util.sse(numpy.array(sim_values_sse), numpy.array(exp_values_sse)), len(sim_values_sse), 
        time_center, numpy.array(sim_values_sse), numpy.array(exp_values_sse), [1.0 - i for i in scores])

def find_bounds(times, values):
    "find the maximum amount left and right the system can be rolled based on 10% of peak max"
    peak_max = max(values)
    cutoff = 0.1 * peak_max

    data = (values > cutoff).nonzero()

    min_index = data[0][0]
    max_index = data[0][-1]

    return [-min_index, len(values) - max_index - 1]

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol):
    temp = {}
    data = pandas.read_csv(feature['fraction_csv'])
    rows, cols = data.shape

    headers = data.columns.values.tolist()

    start = numpy.array(data.iloc[:, 0])
    stop = numpy.array(data.iloc[:, 1])

    smallestTime = min(start - stop)
    abstolFraction = abstol * smallestTime

    funcs = []

    for idx, component in enumerate(headers[2:], 2):
        value = numpy.array(data.iloc[:, idx])
        funcs.append((int(component), score.value_function(max(value), abstolFraction)))

    settings.count = 3 * len(funcs)

    temp['data'] = data
    temp['start'] = start
    temp['stop'] = stop
    temp['timeFunc'] = score.time_function(CV_time, 0, diff_input = True)
    temp['components'] = [int(i) for i in headers[2:]]
    temp['samplesPerComponent'] = rows
    temp['CV_time'] = CV_time
    temp['funcs'] = funcs
    temp['unit'] = feature['unit_name']
    temp['peak_max'] = data.iloc[:,2:].max().min()
    return temp

def headers(experimentName, feature):
    data = pandas.read_csv(feature['fraction_csv'])
    rows, cols = data.shape

    data_headers = data.columns.values.tolist()

    temp = []
    for component in data_headers[2:]:
        temp.append('%s_%s_Component_%s_Similarity' % (experimentName, feature['name'], component))
        temp.append('%s_%s_Component_%s_Time' % (experimentName, feature['name'], component))
        temp.append('%s_%s_Component_%s_Value' % (experimentName, feature['name'], component))
    return temp





import CADETMatch.util as util
import numpy

name = "norm_volume_area"
count = 2
count_extended = 3

def getUnit(location):
    return location.split('/')[3]

def transform(parameter):
    minVolume = parameter['minVolume']
    maxVolume = parameter['maxVolume']
    minArea = parameter['minArea']
    maxArea = parameter['maxArea']

    def trans_volume(i):
        return (i - minVolume)/(maxVolume - minVolume)

    def trans_area(i):
        return (i - minArea)/(maxArea - minArea)

    return [trans_volume, trans_area]

def untransform(seq, cache, parameter):
    minVolume = parameter['minVolume']
    maxVolume = parameter['maxVolume']
    minArea = parameter['minArea']
    maxArea = parameter['maxArea']

    minValues = numpy.array([minVolume, minArea])
    maxValues = numpy.array([maxVolume, maxArea])

    values = numpy.array(seq)

    values = (maxValues - minValues) * values + minValues

    values = [values[1], values[0]/values[1]]
    headerValues = [values[0], values[1], values[0]*values[1]]
    return values, headerValues

def untransform_matrix(matrix, cache, parameter):
    temp = untransform_matrix_inputorder(matrix, cache, parameter)
    values = numpy.zeros(temp.shape)
    values[:,0] = temp[:,1]
    values[:,1] = temp[:,0] / temp[:,1]
    return values

def untransform_matrix_inputorder(matrix, cache, parameter):
    minVolume = parameter['minVolume']
    maxVolume = parameter['maxVolume']
    minArea = parameter['minArea']
    maxArea = parameter['maxArea']

    minValues = numpy.array([minVolume, minArea])
    maxValues = numpy.array([maxVolume, maxArea])

    temp = (maxValues - minValues) * matrix + minValues

    values = numpy.zeros(temp.shape)
    values[:,0] = temp[:,0]
    values[:,1] = temp[:,1]

    return values

def setSimulation(sim, parameter, seq, cache, experiment):
    values, headerValues = untransform(seq, cache, parameter)
    
    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        area_location = parameter['area_location']
        length_location = parameter['length_location']
    
        sim[area_location.lower()] = values[0]
        sim[length_location.lower()] = values[1]

    return values, headerValues

def setupTarget(parameter):
    area_location = parameter['area_location']
    length_location = parameter['length_location']
    bound = parameter['bound']
    comp = parameter['component']

    sensitivityOk = 1
    nameArea = area_location.rsplit('/', 1)[-1]
    nameLength = length_location.rsplit('/', 1)[-1]
    unit = int(area_location.split('/')[3].replace('unit_', ''))

    return [(nameArea, unit, comp, bound), (nameLength, unit, comp, bound)], sensitivityOk

def getBounds(parameter):
    return [0.0, 0.0], [1.0, 1.0]

def getHeaders(parameter):
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("Area Comp:%s Bound:%s" % (comp, bound))
    headers.append("Length Comp:%s Bound:%s" % (comp, bound))
    headers.append("Volume Comp:%s Bound:%s" % (comp, bound))
    return headers

def getHeadersActual(parameter):
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("Volume Comp:%s Bound:%s" % (comp, bound))
    headers.append("Area Comp:%s Bound:%s" % (comp, bound))
    return headers

def setBounds(parameter, lb, ub):
    parameter['minVolume'] = lb[2]
    parameter['maxVolume'] = ub[2]
    parameter['minArea'] = lb[0]
    parameter['maxArea'] = ub[0]


import CADETMatch.util as util
import numpy

name = "norm_nu_sigma"
count = 2
count_extended = 3

def getUnit(location):
    return location.split('/')[3]

def untransform(seq, cache, parameter):
    minNu = parameter['minNu']
    maxNu = parameter['maxNu']
    minSigma = parameter['minSigma']
    maxSigma = parameter['maxSigma']

    minValues = numpy.array([minNu, minNu+minSigma])
    maxValues = numpy.array([maxNu, maxNu+maxSigma])

    values = numpy.array(seq)

    values = (maxValues - minValues) * values + minValues

    values = [values[0], values[1] - values[0]]
    headerValues = [values[0], values[1], values[0]+values[1]]
    return values, headerValues

def untransform_matrix(matrix, cache, parameter):
    values = untransform_matrix_inputorder(matrix, cache, parameter)
    values[:,1] = values[:,1] - values[:,0]
    return values

def untransform_matrix_inputorder(matrix, cache, parameter):
    minNu = parameter['minNu']
    maxNu = parameter['maxNu']
    minSigma = parameter['minSigma']
    maxSigma = parameter['maxSigma']

    minValues = numpy.array([minNu, minNu+minSigma])
    maxValues = numpy.array([maxNu, maxNu+maxSigma])

    temp = (maxValues - minValues) * matrix + minValues

    values = numpy.zeros(temp.shape)
    values[:,0] = temp[:,0]
    values[:,1] = temp[:,1]

    return values

def setSimulation(sim, parameter, seq, cache, experiment):
    values, headerValues = untransform(seq, cache, parameter)
    
    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        nu_location = parameter['nu_location']
        sigma_location = parameter['sigma_location']
    
        comp = parameter['component']
        bound = parameter['bound']
    
        unit = getUnit(nu_location)
        boundOffset = util.getBoundOffset(sim.root.input.model[unit])

        position = boundOffset[comp] + bound
        sim[nu_location.lower()][position] = values[0]
        sim[sigma_location.lower()][position] = values[1]

    return values, headerValues

def setupTarget(parameter):
    nu_location = parameter['nu_location']
    sigma_location = parameter['sigma_location']
    bound = parameter['bound']
    comp = parameter['component']

    sensitivityOk = 1
    nameNu = nu_location.rsplit('/', 1)[-1]
    nameSigma = sigma_location.rsplit('/', 1)[-1]
    unit = int(nu_location.split('/')[3].replace('unit_', ''))

    return [(nameNu, unit, comp, bound), (nameSigma, unit, comp, bound)], sensitivityOk

def getBounds(parameter):
    return [0.0, 0.0], [1.0, 1.0]

def getHeaders(parameter):
    nu_location = parameter['nu_location']
    sigma_location = parameter['sigma_location']
    nameNu = nu_location.rsplit('/', 1)[-1]
    nameSigma = sigma_location.rsplit('/', 1)[-1]
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("%s Comp:%s Bound:%s" % (nameNu, comp, bound))
    headers.append("%s Comp:%s Bound:%s" % (nameSigma, comp, bound))
    headers.append("%s+%s Comp:%s Bound:%s" % (nameNu, nameSigma, comp, bound))
    return headers

def getHeadersActual(parameter):
    nu_location = parameter['nu_location']
    sigma_location = parameter['sigma_location']
    nameNu = nu_location.rsplit('/', 1)[-1]
    nameSigma = sigma_location.rsplit('/', 1)[-1]
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("%s Comp:%s Bound:%s" % (nameNu, comp, bound))
    headers.append("%s+%s Comp:%s Bound:%s" % (nameNu, nameSigma, comp, bound))
    return headers

def setBounds(parameter, lb, ub):
    parameter['minNu'] = lb[0]
    parameter['maxNu'] = ub[0]
    parameter['minSigma'] = lb[1]
    parameter['maxSigma'] = ub[1]


import CADETMatch.util as util
import numpy
import CADETMatch.calc_coeff as calc_coeff

name = "linear"
count = 2
count_extended = 2

def getUnit(location):
    return location.split('/')[3]

def transform(parameter):
    def trans_a(i):
        return i

    def trans_b(i):
        return i

    return [trans_a, trans_b]

def untransform(seq, cache, parameter):
    values = [seq[0], seq[1]]
    headerValues = values
    return values, headerValues

def untransform_matrix(matrix, cache, parameter):
    values = numpy.array(matrix)
    return values

untransform_matrix_inputorder = untransform_matrix

def setSimulation(sim, parameter, seq, cache, experiment):
    values, headerValues = untransform(seq, cache, parameter)

    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        location = parameter['location']
    
        minX = parameter['minX']
        maxX = parameter['maxX']

        x_name = parameter['x_name']

        x_value = experiment[x_name]

        slope, intercept = calc_coeff.linear_coeff(minX, values[0], maxX, values[1])

        value = calc_coeff.linear(x_value, slope, intercept)

        try:
            comp = parameter['component']
            bound = parameter['bound']
            index = None
        except KeyError:
            index = parameter['index']
            bound = None

        if bound is not None:
            unit = getUnit(location)
            boundOffset = util.getBoundOffset(sim.root.input.model[unit])

            if comp == -1:
                position = ()
                sim[location.lower()] = value
            else:
                position = boundOffset[comp] + bound
                sim[location.lower()][position] = value

        if index is not None:
            sim[location.lower()][index] = value

    return values, headerValues

def setupTarget(parameter):
    location = parameter['location']
    bound = parameter['bound']
    comp = parameter['component']

    name = location.rsplit('/', 1)[-1]
    sensitivityOk = 1

    try:
        unit = int(location.split('/')[3].replace('unit_', ''))
    except ValueError:
        unit = ''
        sensitivityOk = 0

    return [(name, unit, comp, bound),], sensitivityOk

def getBounds(parameter):
    minLower = parameter['minLower']
    maxLower = parameter['maxLower']
    minUpper = parameter['minUpper']
    maxUpper = parameter['maxUpper']
    
    minValues = numpy.array([minLower, minUpper])
    maxValues = numpy.array([maxLower, maxUpper])

    return minValues, maxValues

def getHeaders(parameter):
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("Lower Comp:%s Bound:%s" % (comp, bound))
    headers.append("Upper Comp:%s Bound:%s" % (comp, bound))
    return headers

def getHeadersActual(parameter):
    return getHeaders(parameter)

def setBounds(parameter, lb, ub):
    parameter['minLower'] = lb[0]
    parameter['maxLower'] = ub[0]
    parameter['minUpper'] = lb[2]
    parameter['maxUpper'] = ub[2]





import scipy.stats
import numpy
import scipy.optimize
import scipy.interpolate
import scipy.signal
import numpy.linalg
import CADETMatch.calc_coeff as calc_coeff
import scoop

def roll(x, shift):
    if shift > 0:
        temp = numpy.pad(x,(shift,0), mode='constant')
        return temp[:-shift]
    elif shift < 0:
        temp = numpy.pad(x,(0,numpy.abs(shift)), mode='constant')
        return temp[numpy.abs(shift):]
    else:
        return x


def roll_spline(times, values, shift):
    "this function does approximately what the roll function does except that is used the spline so the shift does not have to be an integer and the points are resampled"

    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, values, ext=1)

    times_new = times - shift

    values_new = spline(times_new)

    return values_new


def cross_correlate(exp_time_values, sim_data_values, exp_data_values):
    corr = scipy.signal.correlate(exp_data_values, sim_data_values)/(numpy.linalg.norm(sim_data_values) * numpy.linalg.norm(exp_data_values))

    #need +1 due to how correlate works
    index = numpy.argmax(corr) + 1

    roll_index = index - len(exp_time_values)

    score = corr[index]

    sim_time_values = roll(exp_time_values, shift=int(numpy.ceil(roll_index)))

    diff_time = numpy.abs(exp_time_values[int(len(exp_time_values)/2)] - sim_time_values[int(len(exp_time_values)/2)])

    return score, diff_time

def pearson(exp_time_values, sim_data_values, exp_data_values):
    corr = scipy.signal.correlate(exp_data_values, sim_data_values)/(numpy.linalg.norm(sim_data_values) * numpy.linalg.norm(exp_data_values))

    #need +1 due to how correlate works
    index = numpy.argmax(corr) + 1

    roll_index = index - len(exp_time_values)

    try:
        score = corr[index]
    except IndexError:
        score = 0.0
        scoop.logger.warn("Index error in pearson score at index %s out of %s entries", index, len(corr))

    endTime = exp_time_values[-1]

    sim_time_values = roll(exp_time_values, shift=int(numpy.ceil(roll_index)))

    diff_time = numpy.abs(exp_time_values[int(len(exp_time_values)/2)] - sim_time_values[int(len(exp_time_values)/2)])

    sim_data_values_copy = roll(sim_data_values, shift=int(numpy.ceil(roll_index)))

    pear = scipy.stats.pearsonr(exp_data_values, sim_data_values_copy)
    
    return pear_corr(pear[0]), diff_time

def sse_spline(exp_time_values, sim_data_values, exp_data_values):

    spline = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, sim_data_values, ext=1)

    exp_time_values = numpy.array(exp_time_values)
    exp_data_values = numpy.array(exp_data_values)

    def goal_sse(offset):
        sim_data_values_copy = spline(exp_time_values - offset)

        sse = numpy.sum((exp_data_values - sim_data_values_copy)**2)

        return sse

    #pessimistic bounds for search, find the farthest the peak max can be moved before it rolls off the bounds
    max_index = numpy.argmax(sim_data_values)
    arg_max_time = exp_time_values[max_index]

    min_time = exp_time_values[0] - arg_max_time
    max_time = exp_time_values[-1] - arg_max_time
    
    result_evo = scipy.optimize.differential_evolution(goal_sse, ((min_time, max_time),), polish=True)

    diff_time = -result_evo.x[0]

    sse = result_evo.fun

    return sse, diff_time

def pearson_spline(exp_time_values, sim_data_values, exp_data_values):

    spline = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, sim_data_values, ext=1)

    exp_time_values = numpy.array(exp_time_values)
    exp_data_values = numpy.array(exp_data_values)

    def goal_pearson(offset):
        sim_data_values_copy = spline(exp_time_values - offset)

        pear = scipy.stats.pearsonr(exp_data_values, sim_data_values_copy)

        return -pear[0]

    #pessimistic bounds for search, find the farthest the peak max can be moved before it rolls off the bounds
    max_index = numpy.argmax(sim_data_values)
    arg_max_time = exp_time_values[max_index]

    min_time = exp_time_values[0] - arg_max_time
    max_time = exp_time_values[-1] - arg_max_time
    
    result_evo = scipy.optimize.differential_evolution(goal_pearson, ((min_time, max_time),), polish=True)

    diff_time = -result_evo.x[0]

    pear = -result_evo.fun


    #sse test stuff
    sim_data_values_copy = spline(exp_time_values - result_evo.x[0])
    sse = numpy.sum((exp_data_values - sim_data_values_copy)**2)

    return pear_corr(pear), diff_time

def time_function_decay(CV_time, peak_time, diff_input=False):
    x_exp = numpy.array([0, 1.0*CV_time])
    y_exp = numpy.array([1, 0.5])

    #a, b = calc_coeff.exponential_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])

    a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    
    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        #value = max(0.0, calc_coeff.exponential(diff, a, b))
        value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper

def time_function_decay_cv(CV_time, times, peak_time):

    time = numpy.max(numpy.abs(numpy.array([times[-1] - peak_time, peak_time - times[0]])))
    cv_time = time /CV_time
    
    x_exp = numpy.array([0, cv_time])
    y_exp = numpy.array([1, 0.0])

    a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    
    def wrapper(x):
        diff = x/CV_time
        value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper

def time_function_cv(CV_time, times, peak_time):

    time = numpy.max(numpy.abs(numpy.array([times[-1] - peak_time, peak_time - times[0]])))
    cv_time = time /CV_time

    x_lin_1 = numpy.array([0.0, 0.1])
    y_lin_1 = numpy.array([1.0, 0.95])
    
    x_lin_2 = numpy.array([0.1, cv_time])
    y_lin_2 = numpy.array([0.95, 0.0])

    a1, b1 = calc_coeff.linear_coeff(x_lin_1[0], y_lin_1[0], x_lin_1[1], y_lin_1[1])
    a2, b2 = calc_coeff.linear_coeff(x_lin_2[0], y_lin_2[0], x_lin_2[1], y_lin_2[1])
    
    def wrapper(x):
        diff_cv = x/CV_time
        if diff_cv <= 0.1:
            value = max(0.0, calc_coeff.linear(diff_cv, a1, b1))
        else:
            value = max(0.0, calc_coeff.linear(diff_cv, a2, b2))
        return value
    return wrapper


def time_function_decay_exp(CV_time, peak_time, diff_input=False):
    x_exp = numpy.array([0, 1.0*CV_time])
    y_exp = numpy.array([1, 0.5])

    a, b = calc_coeff.exponential_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])

    #a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    
    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        value = max(0.0, calc_coeff.exponential(diff, a, b))
        #value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper

def time_function(CV_time, peak_time, diff_input=False):
    x_lin = numpy.array([0, 4*CV_time])
    #y_lin = numpy.array([1, 0.0])

    y_lin = numpy.array([1, 0.05])

    a, b = calc_coeff.exponential_coeff(x_lin[0], y_lin[0], x_lin[1], y_lin[1])
    #a, b = calc_coeff.linear_coeff(x_lin[0], y_lin[0], x_lin[1], y_lin[1])

    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        #if diff < CV_time/2.0:
        #value = max(0.0, calc_coeff.linear(diff, a, b))
        value = max(0.0, calc_coeff.exponential(diff, a, b))

        return value

    return wrapper

def time_function2(CV_time, peak_time, diff_input=False):
    #x_exp = numpy.array([CV_time/2.0, 10.0*CV_time])
    x_exp = numpy.array([60.0, 2.0*CV_time])
    y_exp = numpy.array([0.97, 0.5])

    #x_lin = numpy.array([0, CV_time/2.0])
    x_lin = numpy.array([0, 60.0])
    y_lin = numpy.array([1, 0.97])

    a_exp, b_exp = calc_coeff.exponential_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    a_lin, b_lin = calc_coeff.linear_coeff(x_lin[0], y_lin[0], x_lin[1], y_lin[1])

    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        #if diff < CV_time/2.0:
        if diff < 60.0:
            value = calc_coeff.linear(diff, a_lin, b_lin)
        else:
            value = calc_coeff.exponential(diff, a_exp, b_exp)

        return value

    return wrapper

def value_function(peak_height, tolerance=1e-8, bottom_score = 0.0):
    #if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 1.0])
    y = numpy.array([1.0, bottom_score])

    #a, b = calc_coeff.exponential_coeff(x[0], y[0], x[1], y[1])
    a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])
    
    if numpy.abs(peak_height) < tolerance:
        scoop.logger.warn("peak height less than tolerance %s %s", tolerance, peak_height)
        def wrapper(x):
            if numpy.abs(x) < tolerance:
                return 1.0
            else:
                diff = numpy.abs(x-tolerance)/numpy.abs(tolerance)
                #return max(0, calc_coeff.exponential(diff, a, b))
                return max(0, calc_coeff.linear(diff, a, b))
    else:
        def wrapper(x):
            diff = numpy.abs(x-peak_height)/numpy.abs(peak_height)
            #return max(0, calc_coeff.exponential(diff, a, b))
            return max(0, calc_coeff.linear(diff, a, b))

    return wrapper

def value_function_exp(peak_height, tolerance=1e-8, bottom_score = 0.05):
    #if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 1.0])
    y = numpy.array([1.0, bottom_score])

    a, b = calc_coeff.exponential_coeff(x[0], y[0], x[1], y[1])
    #a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])
    
    if numpy.abs(peak_height) < tolerance:
        scoop.logger.warn("peak height less than tolerance %s %s", tolerance, peak_height)
        def wrapper(x):
            if numpy.abs(x) < tolerance:
                return 1.0
            else:
                diff = numpy.abs(x-tolerance)/numpy.abs(tolerance)
                return max(0, calc_coeff.exponential(diff, a, b))
                #return max(0, calc_coeff.linear(diff, a, b))
    else:
        def wrapper(x):
            diff = numpy.abs(x-peak_height)/numpy.abs(peak_height)
            return max(0, calc_coeff.exponential(diff, a, b))
            #return max(0, calc_coeff.linear(diff, a, b))

    return wrapper

def slope_function(peak_slope):
    #if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 4.0])
    y = numpy.array([1.0, 0.0])

    a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])
    
    def wrapper(x):
        diff = numpy.abs(x-peak_slope)/numpy.abs(peak_slope)
        return max(0, calc_coeff.linear(diff, a, b))

    return wrapper

#def pear_corr(cr):
#    #handle the case where a nan is returned
#    if numpy.isnan(cr):
#        return 0.0
#    if cr < 0.5:
#        out = 1.0/3.0 * cr + 1.0/3.0
#    else:
#        out = cr
#    return out

#def pear_corr(cr):
#    #handle the case where a nan is returned
#    if numpy.isnan(cr):
#        return 0.0
#    if cr < 0.7:
#        a,b = calc_coeff.linear_coeff(-1, 0, 0.7, 0.5)
#        out = calc_coeff.linear(cr, a, b)
#    else:
#        a,b = calc_coeff.linear_coeff(0.7, 0.5, 1.0, 1.0)
#        out = calc_coeff.linear(cr, a, b)
#    return out

def pear_corr(cr):
    #handle the case where a nan is returned
    if numpy.isnan(cr):
        return 0.0
    return (1+cr)/2


import CADETMatch.util as util
import CADETMatch.score as score
import numpy
import pandas
from addict import Dict
import sys

"""DO NOT USE THIS SCORE. This score only exists for the purpose of a paper to confirm that
this is not an appropriate method to solve fractionation problems.
"""

name = "fractionationSSE"
settings = Dict()
settings.adaptive = False
settings.badScore = -sys.float_info.max
settings.meta_mask = True
settings.count = 1
settings.failure = [0.0] * settings.count, 1e6, 1, numpy.array([0.0]), numpy.array([0.0]), numpy.array([1e6]), [1.0] * settings.count

def run(sim_data, feature):
    "similarity, value, start stop"
    simulation = sim_data['simulation']
    start = feature['start']
    stop = feature['stop']
    comps = feature['comps']
    data = feature['data']

    time_center = (start + stop)/2.0

    times = simulation.root.output.solution.solution_times

    sim_values_sse = []
    exp_values_sse = []
   
    graph_sim = {}
    graph_exp = {}

    for component  in comps:
        exp_values = numpy.array(data[str(component)])
        sim_value = simulation.root.output.solution[feature['unit']]["solution_outlet_comp_%03d" % component]

        fractions = util.fractionate(start, stop, times, sim_value)

        exp_values_sse.extend(exp_values)
        sim_values_sse.extend(fractions)

        graph_sim[component] = list(zip(time_center, fractions))
        graph_exp[component] = list(zip(time_center, exp_values))

    #sort lists
    for key, value in graph_sim.items():
        value.sort()
    for key, value in graph_exp.items():
        value.sort()

    sim_data['graph_exp'] = graph_exp
    sim_data['graph_sim'] = graph_sim

    sse = util.sse(numpy.array(sim_values_sse), numpy.array(exp_values_sse))

    return [-sse,], sse, len(sim_values_sse), time_center, numpy.array(sim_values_sse), numpy.array(exp_values_sse), [sse,]

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol):
    temp = {}
    data = pandas.read_csv(feature['fraction_csv'])
    rows, cols = data.shape
    
    #funcs = []
    #headers = data.columns.values.tolist()

    #for sample in range(rows):
    #    for component in headers[2:]:
    #        start = data['Start'][sample]
    #        stop = data['Stop'][sample]
    #        value = data[component][sample]

    #         funcs.append( (start, stop, int(component), value) )

    temp['comps'] = [int(i) for i in data.columns.values.tolist()[2:]]

    temp['data'] = data
    temp['start'] = numpy.array(data.iloc[:, 0])
    temp['stop'] = numpy.array(data.iloc[:, 1])
    temp['unit'] = feature['unit_name']

    smallestTime = min(data['Stop'] - data['Start'])
    abstolFraction = abstol * smallestTime

    headers = data.columns.values.tolist()
    temp['peak_max'] = data.iloc[:,2:].max().min()
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_SSE" % name]
    return temp




from scipy.optimize._numdiff import approx_derivative
from CADETMatch.cache import cache
from pathlib import Path
import CADETMatch.util as util
import numpy
import scoop

class GradientException(Exception):
    pass

def jac(individual, cache):
    J = approx_derivative(fitness_grad, individual, method='3-point', kwargs={'cache':cache})
    cond = numpy.linalg.cond(J)
    u, s, v = numpy.linalg.svd(J)
    minSing = numpy.min(s)
    maxSing = numpy.max(s)
    scoop.logger.info('%s has condition %s  min sing:  %s   max sing: %s', individual, cond, minSing, maxSing)
    return J

def fitness_grad(individual, cache):
    scores = []
    error = 0.0

    results = {}
    for experiment in cache.settings['experiments']:
        result = runExperiment(individual, experiment, cache.settings, cache.target, cache)
        if result is not None:
            results[experiment['name']] = result
            scores.extend(results[experiment['name']]['scores'])
            error += results[experiment['name']]['error']
        else:
            raise GradientException("Gradient caused simulation failure, aborting")
   
    return scores

def runExperiment(individual, experiment, settings, target, cache):
    if 'simulation' not in experiment:
        templatePath = Path(settings['resultsDirMisc'], "template_%s.h5" % experiment['name'])
        templateSim = Cadet()
        templateSim.filename = templatePath.as_posix()
        templateSim.load()
        experiment['simulation'] = templateSim

    return util.runExperiment(individual, experiment, settings, target, experiment['simulation'], experiment['simulation'].root.timeout, cache)


#This is a simple implemtnation of NSGA2 that runs on a single thread and has the same kind of interface as the scipy optimize tools
#The reason for this is that finding the maximum likelihood and bandwidth works better with NSGA2 and the optimizers built into scipy

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
import SALib.sample.sobol_sequence
import numpy
import random
import array
import scoop

def sobolGenerator(icls, dimension, lb, ub, n):
    ub = numpy.array(ub)
    lb = numpy.array(lb)
    if n > 0:
        populationDimension = dimension
        populationSize = n
        sobol = SALib.sample.sobol_sequence.sample(populationSize, populationDimension)
        data = numpy.apply_along_axis(list, 1, sobol) * (ub-lb) + lb
        data = list(map(icls, data))
        return data
    else:
        return []

def nsga2_min(func, lb, ub, ngen=8000, mu=200, cxpb=0.8, args=None, stop=40):
    FitnessMin = create("FitnessMin", base.Fitness, weights=(-1.0,))
    Individual = create("Individual", list, fitness=FitnessMin)

    toolbox = base.Toolbox()

    toolbox.register("evaluate", func, *args)
    toolbox.register("population", sobolGenerator, Individual, len(lb), lb, ub)
    toolbox.register("mate", tools.cxSimulatedBinaryBounded, low=lb, up=ub, eta=2.0)
    toolbox.register("mutate", tools.mutPolynomialBounded, low=lb, up=ub, eta=1.0, indpb=1.0/len(lb))
    toolbox.register("select", tools.selNSGA2)

    toolbox.register('map', map)

    pop = toolbox.population(n=mu)

    best = []
    best_score = 1e308
    last_progress = 0

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in pop if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit
        if fit[0] < best_score:
            best_score = fit[0]
            best = ind

    # This is just to assign the crowding distance to the individuals
    # no actual selection is done
    pop = toolbox.select(pop, len(pop))

    # Begin the generational process
    for gen in range(1, ngen):
        # Vary the population
        offspring = tools.selTournamentDCD(pop, len(pop))
        offspring = [toolbox.clone(ind) for ind in offspring]
        
        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if random.random() <= cxpb:
                toolbox.mate(ind1, ind2)
            
            toolbox.mutate(ind1)
            toolbox.mutate(ind2)
            del ind1.fitness.values, ind2.fitness.values
        
        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

            if fit[0] < best_score:
                best_score = fit[0]
                best = ind
                last_progress = gen


        if gen - last_progress > stop:
            scoop.logger.info("stopped at gen %s", gen)
            break

        # Select the next generation population
        pop = toolbox.select(pop + offspring, mu)

        scoop.logger.info('best_score\t %s \tbest\t %s', best_score, best)
        
    result = {}
    result['x'] = best
    result['fun'] = best_score
    result['pop'] = pop

    return result


#From DEAP and modified, need to get rid of this entirely and figure out how to just build the class
#I modified this version so it does not create a global class
def create(name, base, **kargs):
    dict_inst = {}
    dict_cls = {}
    for obj_name, obj in kargs.items():
        if isinstance(obj, type):
            dict_inst[obj_name] = obj
        else:
            dict_cls[obj_name] = obj

    def initType(self, *args, **kargs):
        """Replace the __init__ function of the new type, in order to
        add attributes that were defined with **kargs to the instance.
        """
        for obj_name, obj in dict_inst.items():
            setattr(self, obj_name, obj())
        if base.__init__ is not object.__init__:
            base.__init__(self, *args, **kargs)

    objtype = type(str(name), (base,), dict_cls)
    objtype.__init__ = initType
    return objtype


import shutil
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import h5py
import CADETMatch.util as util
from pathlib import Path
import CADETMatch.evo as evo
import scipy.optimize
import numpy
import numpy.linalg
import hashlib
import CADETMatch.score as score
import tempfile
import os
import subprocess
import csv
import time
import scoop
import CADETMatch.cache as cache

class ConditionException(Exception):
    pass

class GradientException(Exception):
    pass


def setupTemplates(cache):
    settings = cache.settings
    target = cache.target
    parms = target['sensitivities']

    for experiment in settings['experiments']:
        HDF5 = experiment['HDF5']
        name = experiment['name']

        simulationSens = Cadet(experiment['simulation'].root)

        template_path_sens = Path(settings['resultsDirMisc'], "template_%s_sens.h5" % name)
        simulationSens.filename = template_path_sens.as_posix()

        simulationSens.root.input.sensitivity.nsens = len(parms)
        simulationSens.root.input.sensitivity.sens_method = 'ad1'

        sensitivity = simulationSens.root.input.sensitivity

        for idx, parm in enumerate(parms):
            name, unit, comp, bound = parm

            paramSection = 'param_%03d' % idx

            sensitivity[paramSection].sens_unit = [unit,]
            sensitivity[paramSection].sens_name = [name,]
            sensitivity[paramSection].sens_comp = [comp,]
            sensitivity[paramSection].sens_reaction = [-1,]
            sensitivity[paramSection].sens_boundphase = [bound,]
            sensitivity[paramSection].sens_section = [-1,]
            sensitivity[paramSection].sens_abstol = [1e-6,]
            sensitivity[paramSection].sens_factor = [1.0,]

        simulationSens.save()
        experiment['simulationSens'] = simulationSens


def search(gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, generation, check_all=False, result_data=None):
    if check_all:
        checkOffspring = offspring
    else:
        checkOffspring = (ind for ind in offspring if util.product_score(ind.fitness.values) > gradCheck)
    checkOffspring = filterOverlapArea(cache, checkOffspring)
    newOffspring = cache.toolbox.map(cache.toolbox.evaluate_grad, checkOffspring)
        
    temp = []
    csv_lines = []
    meta_csv_lines = []

    for i in newOffspring:
        if i is None:
            pass
        elif i.success:
            ind = cache.toolbox.individual_guess(i.x)
            fit, csv_line, results = cache.toolbox.evaluate(ind)

            ind.fitness.values = fit

            csv_line[0] = 'GRAD'

            save_name_base = hashlib.md5(str(list(ind)).encode('utf-8', 'ignore')).hexdigest()

            ind_meta = cache.toolbox.individualMeta(ind)
            ind_meta.fitness.values = csv_line[-4:] #util.calcMetaScores(fit, cache)

            util.update_result_data(cache, ind, fit, result_data, results, csv_line[-4:])

            if csv_line:
                csv_lines.append([time.ctime(), save_name_base] + csv_line)
                onFront = util.updateParetoFront(grad_hof, ind, cache)
                if onFront and not cache.metaResultsOnly:
                    util.processResultsGrad(save_name_base, ind, cache, results)

                onFrontMeta = util.updateParetoFront(meta_hof, ind_meta, cache)
                if onFrontMeta:
                    meta_csv_lines.append([time.ctime(), save_name_base] + csv_line)
                    util.processResultsMeta(save_name_base, ind, cache, results)
                    cache.lastProgressGeneration = generation

            temp.append(ind)
    
    if temp:
        avg, bestMin, bestProd = util.averageFitness(temp, cache)
        if 0.9 * bestProd > gradCheck:
            gradCheck = 0.9 * bestProd

    writer.writerows(csv_lines)
    
    #flush before returning
    csvfile.flush()

    path_meta_csv = cache.settings['resultsDirMeta'] / 'results.csv'
    with path_meta_csv.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)
        writer.writerows(meta_csv_lines)

    util.cleanupFront(cache, None, meta_hof, grad_hof)
    util.writeMetaFront(cache, meta_hof, path_meta_csv)

    return gradCheck, temp

def gradSearch(x, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup(json_path)

    jac_cache = []

    try:
        scoop.logger.info("gradSearch least_squares")

        x = util.convert_individual_grad(x, cache.cache)

        result = scipy.optimize.least_squares(fitness_sens_grad, x, jac=jacobian, method='trf', 
                                               bounds=(cache.cache.MIN_VALUE_GRAD, cache.cache.MAX_VALUE_GRAD), 
                                               gtol=None, ftol=None, xtol=1e-10, x_scale="jac",
                                               kwargs={'jac_cache':jac_cache}, verbose=0)


        scoop.logger.info("start: %s  stop: %s distance: %s  sse: %s", x, result.x, numpy.linalg.norm(x - result.x), numpy.sum(result.fun**2))
        #scoop.logger.info("result %s", result)
        result.x = util.convert_individual_inverse_grad(result.x, cache.cache)
        return result
    except GradientException:
        #If the gradient fails return None as the point so the optimizer can adapt
        scoop.logger.error("Gradient Failure", exc_info=True)
        return None
    except ConditionException:
        scoop.logger.error("Condition Failure", exc_info=True)
        return None

def fitness_sens_grad(individual, jac_cache, finished=0):
    individual = util.convert_individual_inverse_grad(individual, cache.cache)
    return fitness_sens(individual, jac_cache, finished)

def fitness_sens(individual, jac_cache, finished=1):
    minimize = []
    error = 0.0

    diff = []

    jac_temp = []

    results = {}
    for experiment in cache.cache.settings['experiments']:
        result = runExperimentSens(individual, experiment, cache.cache.settings, cache.cache.target, cache.cache)
        if result is not None:
            jac = getJacobianSimulation(result['simulation'], experiment)
            jac_temp.append(jac)
            results[experiment['name']] = result
            minimize.extend(result['minimize'])
            error += result['error']

            #need to get the right diff for SSE
            temp_diff = getDiff(result, experiment)
            diff.extend(temp_diff)
        else:
            raise GradientException("Gradient caused simulation failure, aborting")
   
    

    #jac_cache[tuple(individual)] = numpy.concatenate(jac_temp, 0)
    jac_cache.append(numpy.concatenate(jac_temp, 0))

    #cond = numpy.linalg.cond(jac_cache[tuple(individual)])

    cond = numpy.linalg.cond(jac_cache[0])

    #scoop.logger.info('condition number %s = %s', individual, cond)

    #need to minimize
    diff = numpy.array(diff)
    sse = numpy.sum(diff**2)
    return diff

def getDiff(result, experiment):
    gradsetup = experiment['gradsetup']
    sim = result['simulation']
    solution = sim.root.output.solution
    times = solution.solution_times
    

    sim_value = []
    exp_value = []
    for grad in gradsetup:
        selected_sim = (times >= grad['start']) & (times <= grad['stop'])
        temp = [solution["unit_%03d" % grad['unit']]["solution_outlet_comp_%03d" % comp] for comp in grad['comps']]
        sim_value.append(numpy.sum(numpy.array(temp), axis=0)[selected_sim])

        data = numpy.loadtxt(grad['csv'], delimiter=',')

        time = data[:, 0]
        value = data[:, 1]
        selected_grad = (time >= grad['start']) & (time <= grad['stop'])
        exp_value.append(value[selected_grad])


    sim_value = numpy.concatenate(sim_value, axis=0)
    exp_value = numpy.concatenate(exp_value, axis=0)
    diff_value = sim_value - exp_value
    return diff_value
    

def jacobian(x, jac_cache):
    return jac_cache.pop()
    #x = util.convert_individual_inverse_grad(x, cache.cache)
    #try:
    #    return jac_cache[tuple(x)]
    #except KeyError:
    #    scoop.logger.info('keyerror %s %s', repr(x), [repr(key) for key in jac_cache.keys()])

def saveExperimentsSens(save_name_base, settings, target, results):
    return util.saveExperiments(save_name_base, settings, target, results, settings['resultsDirGrad'], '%s_%s_GRAD.h5')

def runExperimentSens(individual, experiment, settings, target, cache):
    if 'simulationSens' not in experiment:
        templatePath = Path(settings['resultsDirMisc'], "template_%s_sens.h5" % experiment['name'])
        templateSim = Cadet()
        templateSim.filename = templatePath.as_posix()
        templateSim.load()
        experiment['simulationSens'] = templateSim

    return util.runExperiment(individual, experiment, settings, target, experiment['simulationSens'], float(experiment['timeout']), cache)

def getJacobianSimulation(sim, experiment):
    #write out jacobian to jac
    gradsetup = experiment['gradsetup']
    params = len(cache.cache.target['sensitivities'])
    sens = sim.root.output.sensitivity
    times = sim.root.output.solution.solution_times

    jacobians = []
    for grad in gradsetup:
        temp_jac = []
        selected = (times >= grad['start']) & (times <= grad['stop'])
        for idx in range(params):
            temp = [sens["param_%03d" % idx]["unit_%03d" % grad['unit']]["sens_outlet_comp_%03d" % comp] for comp in grad['comps']]
            temp_jac.append(numpy.sum(numpy.array(temp), axis=0))
        jacobian = numpy.squeeze(numpy.array(temp_jac)).T
        jacobian = jacobian[selected,:]
        jacobians.append(jacobian)
    
    jacobian = numpy.concatenate(jacobians, axis=0)
    return jacobian

def filterOverlapArea(cache, checkOffspring, cutoff=0.01):
    """if there is no overlap between the simulation and the data there is no gradient to follow and these entries need to be skipped
    This only applies if the score is SSE or gradVector is True"""
    temp = cache.toolbox.map(cache.toolbox.evaluate, map(list, checkOffspring))

    temp_offspring = []

    for ind, (fit, csv_line, results) in zip(map(list, checkOffspring), temp):
        temp_area_total = 0.0
        temp_area_overlap = 0.0
        if results is not None:
            for exp in results.values():
                sim_times = exp['sim_time']
                sim_values = exp['sim_value']
                exp_values = exp['exp_value']

                for sim_time, sim_value, exp_value in zip(sim_times, sim_values, exp_values):
                    temp_area_total += numpy.trapz(exp_value, sim_time)
                    temp_area_overlap += numpy.trapz(numpy.min([sim_value, exp_value], 0), sim_time)

            percent = temp_area_overlap / temp_area_total
            if percent > cutoff:
                temp_offspring.append(ind)
            else:
                scoop.logger.info('removed %s for insufficient overlap in gradient descent', ind)
        else:
            scoop.logger.info('removed %s for failure', ind)

    if checkOffspring:
        scoop.logger.info("overlap okay offspring %s", temp_offspring)
    return temp_offspring



import importlib
import argparse
import sys
import subprocess
import psutil

def makeParser():
    """Create the CADETMatch module arguments parser."""
    parser = argparse.ArgumentParser(
        description="Starts a parallel version of match using SCOOP.",
        prog="{0} -m CADETMatch.scoop_match".format(sys.executable),
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--json', '-j',
                       help="Path to JSON file",
                       metavar="JSON")
    parser.add_argument('--match',
                        help="Run CADETMatch with Parameter Estimation and Error Modeling",
                        action='store_true')

    parser.add_argument('--generate_corner',
                        help="Generate corner graphs",
                        action='store_true')

    parser.add_argument('--generate_graphs',
                        help="Generate general graphs",
                        action='store_true')

    parser.add_argument('--generate_spearman',
                        help="Generate spearman graphs",
                        action='store_true')
    
    parser.add_argument('--generate_mcmc_plot_tube',
                        help="Generate mcmc plot tube graphs",
                        action='store_true')

    parser.add_argument('--generate_mle',
                        help="Generate maximum liklihood",
                        action='store_true')

    parser.add_argument('-n',
                        help="Number of scoop processes to use. Use 1 for debugging",
                        default=0)

    return parser


def run_command(module, json, number_of_jobs, additional=None):
    command = [sys.executable, '-m', 'scoop']
    if int(number_of_jobs) > 0:
        command.extend(['-n', str(number_of_jobs)])
    #else:
    #    ncpus = psutil.cpu_count(logical=False)
    #    command.extend(['-n', str(ncpus)])

    command.extend([importlib.util.find_spec(module).origin, str(json)])
    if additional is not None:
        command.extend(additional)

    rc = subprocess.run(command, bufsize=1)
    
    return rc.returncode


if __name__ == "__main__":
    parser = makeParser()
    args = parser.parse_args()
    
    if args.match:
        sys.exit(run_command('CADETMatch.match', args.json, args.n))
    if args.generate_corner:
        sys.exit(run_command('CADETMatch.generate_corner_graphs', args.json, args.n))
    if args.generate_graphs:
        sys.exit(run_command('CADETMatch.generate_graphs', args.json, args.n, ['2']))
    if args.generate_spearman:
        sys.exit(run_command('CADETMatch.graph_spearman', args.json, args.n, ['end']))
    if args.generate_mcmc_plot_tube:
        sys.exit(run_command('CADETMatch.mcmc_plot_tube', args.json, args.n))
    if args.generate_mle:
        sys.exit(run_command('CADETMatch.mle', args.json, args.n))



import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import numpy
import numpy.linalg
from addict import Dict
import sys

name = "DextranSSE"
settings = Dict()
settings.adaptive = False
settings.badScore = -sys.float_info.max
settings.meta_mask = True
settings.count = 1
settings.failure = [0.0] * settings.count, 1e6, 1, numpy.array([0.0]), numpy.array([0.0]), numpy.array([1e6]), [1.0] * settings.count

def run(sim_data, feature):
    "special score designed for dextran. This looks at only the front side of the peak up to the maximum slope and pins a value at the elbow in addition to the top"
    exp_time_values = feature['time']
    max_value = feature['max_value']

    selected = feature['selected']
        
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)

    diff = feature['value'] - sim_data_values

    if max(sim_data_values) < max_value: #the system has no point higher than the value we are looking for
        #remove hard failure
        max_value = max(sim_data_values)

    exp_time_values = exp_time_values[selected]
    exp_data_zero = feature['exp_data_zero']

    min_index = numpy.argmax(sim_data_values >= 1e-3*max_value)
    max_index = numpy.argmax(sim_data_values >= max_value)

    sim_data_zero = numpy.zeros(len(sim_data_values))
    sim_data_zero[min_index:max_index+1] = sim_data_values[min_index:max_index+1]

    sse = util.sse(sim_data_zero, exp_data_zero)

    data = [-sse,], sse, len(sim_data_zero), sim_time_values, sim_data_zero, exp_data_zero, [sse,]

    return data

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol):
    temp = {}
    #change the stop point to be where the max positive slope is along the searched interval
    s = util.find_smoothing_factor(selectedTimes, selectedValues)

    exp_spline = util.create_spline(selectedTimes, selectedValues, s).derivative(1)

    values = exp_spline(selectedTimes)
    
    max_index = numpy.argmax(values)
    max_time = selectedTimes[max_index]
    max_value = selectedValues[max_index]

    min_index = numpy.argmax(selectedValues >= 1e-3*max_value)
    min_time = selectedTimes[min_index]
    min_value = selectedValues[min_index]

    exp_data_zero = numpy.zeros(len(selectedValues))
    exp_data_zero[min_index:max_index+1] = selectedValues[min_index:max_index+1]

    exp_spline = util.create_spline(selectedTimes, exp_data_zero, s).derivative(1)

    [high, low] = util.find_peak(selectedTimes, exp_spline(selectedTimes))

    temp['min_time'] = feature['start']
    temp['max_time'] = feature['stop']
    temp['max_value'] = max_value
    temp['exp_data_zero'] = exp_data_zero
    temp['offsetTimeFunction'] = score.time_function_decay_cv(CV_time, selectedTimes, max_time)
    temp['peak_max'] = max_value
    temp['smoothing_factor'] = s
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_SSE" % name,]
    return temp










#This module simulates noise (systemic and random) to use for kernel density estimation based on experimental data

#Currently the randomness is

#Pump flow rate
#Pump delay
#Base noise
#Signal noise

import itertools
import CADETMatch.score as score
from cadet import Cadet, H5
from pathlib import Path
import numpy
import scoop
from sklearn.neighbors.kde import KernelDensity
from sklearn.model_selection import cross_val_score
from sklearn import preprocessing
import copy

bw_tol = 1e-4

import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import h5py

import scipy.optimize

import CADETMatch.util as util
import CADETMatch.cache as cache

from scoop import futures
import CADETMatch.synthetic_error as synthetic_error

import joblib
import subprocess
import sys

def bandwidth_score(bw, data, store):
    bandwidth = 10**bw[0]
    kde_bw = KernelDensity(kernel='gaussian', bandwidth=bandwidth, atol=bw_tol)
    scores = cross_val_score(kde_bw, data, cv=3)
    store.append( [bandwidth, -max(scores)] )
    return -max(scores)

def get_bandwidth(scores, cache):
    store = []

    result = scipy.optimize.differential_evolution(bandwidth_score, bounds = [(-4, 1),], 
                    args = (scores,store,), updating='deferred', workers=futures.map, disp=True,
                    popsize=100)
    bandwidth = 10**result.x[0]
    scoop.logger.info("selected bandwidth %s", bandwidth)

    store = numpy.array(store)
    return bandwidth, store

def mirror(data):
    data_min = numpy.max(data,0) - data
    data_mask = numpy.ma.masked_equal(data_min, 0.0, copy=False)
    min_value = data_mask.min(axis=0)
    
    data_mirror = 1 + 1 - numpy.copy(data) + min_value
    full_data = numpy.vstack([data_mirror, data])
    
    return full_data

def setupKDE(cache):
    scores, simulations = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    bandwidth, store = get_bandwidth(scores_scaler, cache)

    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth, atol=bw_tol).fit(scores_scaler)

    joblib.dump(scaler, mcmcDir / 'kde_scaler.joblib')

    joblib.dump(kde, mcmcDir / 'kde_score.joblib')

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.root.bandwidth = bandwidth
    h5_data.root.store = numpy.array(store)
    h5_data.root.scores = scores
    h5_data.root.scores_mirror = scores_mirror
    h5_data.root.scores_mirror_scaled = scores_scaler
    h5_data.save()

    cwd = str(Path(__file__).parent)
    ret = subprocess.run([sys.executable, 'graph_kde.py', str(cache.json_path),], 
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
    util.log_subprocess('graph_kde.py', ret)

    return kde, scaler

def getScaler(data):
    scaler = preprocessing.StandardScaler().fit(data)
    return scaler

def getKDE(cache):
    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    kde = joblib.load(mcmcDir / 'kde_score.joblib')

    scaler = joblib.load(mcmcDir / 'kde_scaler.joblib')

    return kde, scaler

def getKDEPrevious(cache):
    if 'mcmc_h5' in cache.settings:
        mcmc_h5 = Path(cache.settings['mcmc_h5'])
        mcmcDir = mcmc_h5.parent

        if mcmcDir.exists():
            kde = joblib.load(mcmcDir / 'kde_prior.joblib')

            scaler = joblib.load(mcmcDir / 'kde_prior_scaler.joblib')

            return kde, scaler
    return None, None

def generate_data(cache):
    scores, simulations = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    save_scores = mcmcDir / 'scores_used.npy'

    numpy.save(save_scores, scores)

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    bandwidth, store = get_bandwidth(scores_scaler, cache)

    return scores, bandwidth

def synthetic_error_simulation(json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup(json_path)

    scores = []
    outputs = {}
    simulations = {}
    experiment_failed = False

    for experiment in cache.cache.settings['kde_synthetic']:
        file_path = experiment['file_path']
        delay_settings = experiment['delay']
        flow_settings = experiment['flow']
        load_settings = experiment['load']
        error_slope_settings = experiment['error_slope']
        error_base_settings = experiment['error_base']
        count_settings = experiment['count']
        experimental_csv = experiment['experimental_csv']
        units = experiment['units']
        name = experiment['name']

        dir_base = Path(cache.cache.settings.get('baseDir'))
        file = dir_base / file_path

        data = numpy.loadtxt(experimental_csv, delimiter=',')
        times = data[:,0]

        temp = Cadet()
        temp.filename = file.as_posix()
        temp.load()

        util.setupSimulation(temp, times, cache.cache.target[name]['smallest_peak'], cache.cache)

        nsec = temp.root.input.solver.sections.nsec

        def post_function(simulation):
            error_slope = numpy.random.normal(error_slope_settings[0], error_slope_settings[1], 1)[0]
            base = numpy.max(simulation.root.output.solution.unit_002.solution_outlet_comp_000)/1000.0
            error_base = numpy.random.normal(base, base/error_base_settings, len(simulation.root.output.solution.unit_002.solution_outlet_comp_000))
            simulation.root.output.solution.unit_002.solution_outlet_comp_000 = simulation.root.output.solution.unit_002.solution_outlet_comp_000 * error_slope + error_base
    
        error_delay = Cadet(temp.root)

        delays = numpy.random.uniform(delay_settings[0], delay_settings[1], nsec)
    
        synthetic_error.pump_delay(error_delay, delays)

        flow = numpy.random.normal(flow_settings[0], flow_settings[1], error_delay.root.input.solver.sections.nsec)
    
        synthetic_error.error_flow(error_delay, flow)
    
        load = numpy.random.normal(load_settings[0], load_settings[1], error_delay.root.input.solver.sections.nsec)
    
        synthetic_error.error_load(error_delay, load)


        exp_info = None
        for exp in cache.cache.settings['experiments']:
            if exp['name'] == name:
                exp_info = exp
                break

        result = util.runExperiment(None, exp_info, cache.cache.settings, cache.cache.target, error_delay, error_delay.root.timeout, cache.cache, post_function=post_function)

        if result is not None:
            scores.extend(result['scores'])

            simulations[name] = result['simulation']

            for unit in units:
                outputs['%s_unit_%03d' % (name, int(unit))] = result['simulation'].root.output.solution['unit_%03d' % int(unit)].solution_outlet_comp_000
        else:
            experiment_failed = True

    if experiment_failed:
        return None, None, None
    return scores, simulations, outputs

def generate_synthetic_error(cache):
    if 'kde_synthetic' in cache.settings:
        count_settings = int(cache.settings['kde_synthetic'][0]['count'])
        
        scores_all = []
        simulations_all = {}
        outputs_all = {}

        for scores, simulations, outputs in futures.map(synthetic_error_simulation, [cache.json_path] * count_settings):
            if scores and simulations and outputs:

                scores_all.append(scores)

                for key,value in outputs.items():
                    temp = outputs_all.get(key, [])
                    temp.append(value)
                    outputs_all[key] = temp
                                        
        scores = numpy.array(scores_all)

        times = {}
        for simulation_name, simulation_values in simulations_all.items():
            times[simulation_name] = simulation_values.root.input.solver.user_solution_times
            #Only the first item is needed
            break

        dir_base = cache.settings.get('resultsDirBase')
        file = dir_base / 'kde_data.h5'

        with h5py.File(file, 'w') as hf:
            hf.create_dataset('scores', data=scores)

            for output_name, output in outputs_all.items():
                hf.create_dataset(output_name, data=numpy.array(output))

            for time_name, time in times.items():
                hf.create_dataset(time_name, data=times)
                               
        return scores, simulations_all

    return None, None

def writeVariations(cache, scores, bandwidth, simulations, store):

    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    mcmc_kde = mcmcDir / 'mcmc_kde.h5'

    bandwidth = numpy.array(bandwidth, dtype="f8")

    data, times = getData(simulations)

    store = numpy.array(store)

    with h5py.File(mcmc_kde, 'w') as hf:
        hf.create_dataset("bandwidth", data=bandwidth, maxshape=tuple(None for i in range(bandwidth.ndim)), fillvalue=[0])

        hf.create_dataset("scores", data=scores, maxshape=(None, len(scores[0])))

        hf.create_dataset("bandwidth_scores", data=store, maxshape=(None,2))

        for name, value in times.items():
            hf.create_dataset(name, data=value, maxshape=(None,))

        for name, value in data.items():
            hf.create_dataset(name, data=value, maxshape=(None, len(value[0])))

def getData(variations):

    data = {}
    times = {}

    for variation in variations:
        for experiment in variation:
            name = experiment['name']

            if 'data' in experiment:
                experiment_data = experiment['data']

                key = '%s_%s' % (name, "times")
                if key not in data:
                    times[key] = experiment_data[:,0]

                key = name
                if key not in data:
                    data[key] = []

                data[key].append(experiment_data[:,1])

            for feature in experiment['features']:
                feature_name = feature['name']
                
                if 'data' in feature:
                    experiment_data = feature['data']

                    key = '%s_%s' % (name, feature_name)
                    if key not in data:
                        data[key] = []
                    data[key].append(experiment_data[:,1])

    for key,value in data.items():
        data[key] = numpy.array(value)

    return data, times

def plotVariations(cache, temp):

    data, times = getData(temp)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    for key,value in data.items():
        experiment = key.split('_unit', maxsplit=1)[0]

        time = times['%s_times' % experiment]

        mean = numpy.mean(value, 0)
        std = numpy.std(value, 0)
        minValues = numpy.min(value, 0)
        maxValues = numpy.max(value, 0)

        plt.plot(time, mean)
        plt.fill_between(time, mean - std, mean + std,
                    color='green', alpha=0.2)
        plt.fill_between(time, minValues, maxValues,
                    color='red', alpha=0.2)
        plt.xlabel('time(s)')
        plt.ylabel('conc mol/m^3')
        plt.savefig(str(mcmcDir / ("%s.png" % key ) ), bbox_inches='tight')
        plt.close()

        plt.plot(time, value.transpose(), 'g', alpha=0.04, linewidth=2)
        plt.xlabel('time(s)')
        plt.ylabel('conc mol/m^3')
        plt.savefig(str(mcmcDir / ("%s_probability.png" % key ) ), bbox_inches='tight')
        plt.close()

        plt.plot(time, value.transpose())
        plt.xlabel('time(s)')
        plt.ylabel('conc mol/m^3')
        plt.savefig(str(mcmcDir / ("%s_lines.png" % key ) ), bbox_inches='tight')
        plt.close()

def plotKDE(cache, kde, scores):
    return None

def setupReferenceResult(cache):
    #results = {}
    
    #for experiment in cache.settings['experiments']:
        
    #    templatePath = experiment['reference']
    #    templateSim = Cadet()
    #    templateSim.filename = templatePath.as_posix()
    #    templateSim.run()
    #    templateSim.load()
        
    #    results[experiment['name']] = {'simulation':templateSim}

    temp = []
    for experiment in cache.settings['experiments']:
        temp_exp = dict(experiment)

        if 'csv' in temp_exp:
            temp_exp['data'] = numpy.loadtxt(temp_exp['csv'], delimiter=',')

        for feature in temp_exp['features']:
            if 'csv' in feature:
                feature['data'] = numpy.loadtxt(feature['csv'], delimiter=',')
        temp.append(temp_exp)

    return temp


def mutate(cache, reference_result):
    "generate variations of the data with different types of noise"
    result = []

    for experiment in reference_result:
        experiment_temp = copy.deepcopy(experiment)

        pump_delay_time = None

        if 'data' in experiment_temp:
            data = experiment_temp['data']
            
            pump_delay_time = pump_delay(cache, data, pump_delay_time)

            #pump_flow(cache, data)
            base_noise(cache, data)
            signal_noise(cache, data)

            experiment_temp['data'] = data

        for feature in experiment_temp['features']:
            if 'data' in feature:
                data = feature['data']
            
                pump_delay_time = pump_delay(cache, data, pump_delay_time)
                #pump_flow(cache, data)
                base_noise(cache, data)
                signal_noise(cache, data)
                feature['data'] = data
        
        result.append(experiment_temp)

    return result

def pump_delay(cache, data, pump_delay_time=None):
    "systemic error related to delays in the pump"
    "assume data is mono-spaced in time"
    "time is column 0 and values is column 1"
    times = data[:,0]

    #delay = numpy.random.uniform(0.0, 60.0, 1)
    if pump_delay_time is not None:
        dely = pump_delay_time
    else:
        try:
            pump_mean = cache.settings['kde']['pump_delay_mean']
        except KeyError:
            pump_mean = 0.0

        try:
            pump_std = cache.settings['kde']['pump_delay_std']
        except KeyError:
            pump_std = 1.0
            
        #delay = -1
        #while delay < 0:
        delay = numpy.random.normal(pump_mean, pump_std, 1)

    delay = delay[0]
    #interval = times[1] - times[0]
    #delay = quantize_delay(delay[0], interval)

    #data[:,1] = score.roll(data[:,1], delay)
    data[:,1] = score.roll_spline(data[:,0], data[:,1], delay)

    return delay

def pump_flow(cache, tempSim):
    "random noise related to the pump flow rate"
    "assume 5% margin of error but no other changes to chromatogram"
    "assume change is small enough that the chromatogram shape does not change"
    "can't currently model this without simulation work"
    pass

def base_noise(cache, data):
    "add random noise based on baseline noise"
    "based on looking at experimental data"
    times = data[:,0]
    
    try:
        noise_std = cache.settings['kde']['base_noise_std']
    except KeyError:
        noise_std = 1e-7

    noise = numpy.random.normal(0.0, noise_std, len(times))

    data[:,1] = data[:,1] + noise * max(data[:,1])

def signal_noise(cache, data):
    "add noise to the signal"
    "based on looking at experimental error about +/- .5%"
    times = data[:,0]

    #0.003 base on experiments

    try:
        noise_std = cache.settings['kde']['signal_noise_std']
    except KeyError:
        noise_std = 0.003

    noise = numpy.random.normal(1.0, noise_std, len(times))


    data[:,1] = data[:,1] * noise

def get_outputs(tempSim):
    "get the outputs for tempSim so they can be mutated"
    temp = []
    for unitName, unit in tempSim.root.output.solution.items():
        if "unit" in unitName:
            for solutionName, solution in unit.items():
                if 'outlet' in solutionName:
                    temp.append( ((unitName, solutionName), solution) )
    return temp

def quantize_delay(delay, interval):
    return int(round(delay/float(interval))*float(interval)/interval)

def score_sim(first,second, cache):
    "score first vs second simulation"
    target = {}
    for experiment in first:
        target[experiment["name"]] = cache.setupExperiment(experiment)

    score_sim = []
    diff = 0
    for experiment in second:
        experimentName = experiment['name']

        data = experiment.get('data', None)

        for feature in experiment['features']:
            featureType = feature['type']
            featureName = feature['name']

            if data is None:
                data = feature.get('data', None)

            if featureType in cache.scores:
                scores, sse, sse_count, diff, minimize = cache.scores[featureType].run({'simulation':data}, target[experimentName][featureName])
                score_sim.extend(scores)

    return score_sim